#include "cAlert.h"

#include <ostream>
using std::cout;
using std::endl;


cAlert::cAlert(float* center, float* range, int shapetype, std::string message, OID group, bool positive, bool posedge, bool once, OID fusedelay) {
    vector_cpy(shape.center, center);
    vector_cpy(shape.range, range);
    this->traceable->radius = 1 + fmax(fabs(shape.range[0]), fmax(fabs(shape.range[1]), fabs(shape.range[2])));
    if (this->traceable->radius >= 25) {
        // Make global "position".
        vector_set(traceable->pos, float_NAN, float_NAN, float_NAN);
    } else {
        vector_cpy(traceable->pos, center);
    }
    this->positive = positive;
    this->posedge = posedge;
    this->fusedelay = fusedelay;
    this->once = once;
    this->fired = false;
    this->message = message;
    this->group = group;
    collideable = new rCollideable;
    addRole(COLLIDEABLE, collideable);
    shape.type = shapetype;
    sensitivity.insert(HUMANPLAYER);
    fusedelay = 0;
}


float cAlert::constrainParticle(float* worldpos, float radius, float* localpos, cObject* enactor) {
    //std::cout << "called" << std::endl;
    // Don't react on scanning and shooting.
    if (enactor == NULL) return 0;
    if (once && fired) return 0;
    if (! enactor->allRoles(&sensitivity) ) return 0;

    rShape* s = &shape;
    float depth = 0;
    float projection[3];
    switch (s->type) {
        case rShape::CYLINDER:
        {
            depth = cParticle::constraintParticleByCylinder(worldpos, shape.center, shape.range[2], shape.range[1], projection);
            break;
        }
        case rShape::BOX:
        {
            float min[3];
            float max[3];
            vector_set3i(min, shape.center[i] - shape.range[i]);
            vector_set3i(max, shape.center[i] + shape.range[i]);
            depth = cParticle::constraintParticleByAABB(worldpos, min, max, projection);
            break;
        }
        case rShape::SPHERE:
        {
            depth = cParticle::constraintParticleBySphere(worldpos, shape.center, shape.range[0], projection);
            break;
        }
        default:
            break;
    }
    //std::cout << depth << std::endl;
    // Is that possible intruder yet unknown.
    if (intruders.find(enactor->base->oid) == intruders.end()) {
        // The intruder was outside and now there is an intrusion.
        if ((depth != 0) == positive) {
            intruders.insert(enactor->base->oid);
            if (posedge) {
                fired = true;
                //cout << "fusedelay=" << fusedelay << endl;
                cWorld::instance->sendMessageT(fusedelay, base->oid, group, message, NULL);
            }
        }
    } else  {
        // The intruder was inside and now the intruder left the zone.
        if ((depth == 0) == positive) {
            intruders.erase(enactor->base->oid);
            if (!posedge) {
                fired = true;
                cWorld::instance->sendMessageT(fusedelay, base->oid, group, message, NULL);
            }
        }
    }
    //if (fired) cout << "FIRED" << endl;
    return 0;
}


void cAlert::drawEffect() {
    return;
    glPushMatrix();
    {
        glPushAttrib(GL_ALL_ATTRIB_BITS);
        {
            glDisable(GL_CULL_FACE);
            rShape* s = &shape;
            float* center = s->center;
            float* range = s->range;
            switch (s->type) {
                case rShape::CYLINDER:
                {
                    break;
                }
                case rShape::BOX:
                {
                    glBegin(GL_QUADS);
                    glColor4f(1,1,1,0.9);
                    glVertex3f(center[0]-range[0],center[1]-range[1],center[2]-range[2]);
                    glVertex3f(center[0]+range[0],center[1]-range[1],center[2]-range[2]);
                    glColor4f(1,0,0,0.1);
                    glVertex3f(center[0]+range[0],center[1]+range[1],center[2]-range[2]);
                    glVertex3f(center[0]-range[0],center[1]+range[1],center[2]-range[2]);
                    glColor4f(1,1,1,0.9);
                    glVertex3f(center[0]-range[0],center[1]-range[1],center[2]+range[2]);
                    glVertex3f(center[0]+range[0],center[1]-range[1],center[2]+range[2]);
                    glColor4f(1,0,0,0.1);
                    glVertex3f(center[0]+range[0],center[1]+range[1],center[2]+range[2]);
                    glVertex3f(center[0]-range[0],center[1]+range[1],center[2]+range[2]);

                    glColor4f(1,1,1,0.9);
                    glVertex3f(center[0]-range[0],center[1]-range[1],center[2]-range[2]);
                    glVertex3f(center[0]-range[0],center[1]-range[1],center[2]+range[2]);
                    glColor4f(1,0,0,0.1);
                    glVertex3f(center[0]-range[0],center[1]+range[1],center[2]+range[2]);
                    glVertex3f(center[0]-range[0],center[1]+range[1],center[2]-range[2]);
                    glColor4f(1,1,1,0.9);
                    glVertex3f(center[0]+range[0],center[1]-range[1],center[2]-range[2]);
                    glVertex3f(center[0]+range[0],center[1]-range[1],center[2]+range[2]);
                    glColor4f(1,0,0,0.1);
                    glVertex3f(center[0]+range[0],center[1]+range[1],center[2]+range[2]);
                    glVertex3f(center[0]+range[0],center[1]+range[1],center[2]-range[2]);
                    glEnd();
                    break;
                }
                case rShape::SPHERE:
                {
                    break;
                }
                default:
                    break;
            }
    
        }
        glPopAttrib();
    }
    glPopMatrix();
}
