// cBackground.cpp

#include "cBackground.h"

#include <GL/gl.h>

#include <cstdio>
#include <memory.h>
#include <stdlib.h>

#include <iostream>
using std::cout;
using std::endl;

#include "psi3d/macros.h"

#define DETAIL 0

extern unsigned char* loadTGA(const char *fname, int *w, int* h, int* bpp);

float nm_to_R(int nm) {
    if (nm <= 390) return 1.0f;
    if (nm >= 440 && nm <= 510) return 0.0f;
    if (nm >= 580) return 1.0f;
    if (nm < 440) return (float(nm - 390) / float(440 - 390));
    return (float(nm - 510) / float(580 - 510));
}

float nm_to_G(int nm) {
    if (nm <= 440) return 0.0f;
    if (nm >= 640) return 1.0f;
    if (nm >= 490 && nm <= 580) return 1.0f;
    if (nm < 490) return (float(nm - 440) / float(490 - 440));
    return (float(nm - 580) / float(640 - 580));
}

float nm_to_B(int nm) {
    if (nm < 490) return 1.0f;
    if (nm > 510) return 0.0f;
    return (float(nm - 490) / float(510 - 490));
}

enum Texids {
    T_ICE, T_MOUNTAIN, T_EARTH,
    T_LAVOS, T_MOON, T_SUN, T_CLOUDS,
    T_GALAXY_NEGX, T_GALAXY_POSX,
    T_GALAXY_NEGY, T_GALAXY_POSY,
    T_GALAXY_NEGZ, T_GALAXY_POSZ,
};


#include "psi3d/noise.h"

static void getNebula(float* v3f, float* c4f) {

    float c[4];
    loopj(3) {
        unsigned char seed = 31+j*17;
        float sum = 0;
        float fsum = 0;
        for (int i = 1; i < 7; i++) {
            float p = (1 << i)*0.5;
            float f = (8 - i)*4;
            sum += f * simplex3(v3f[0]*p, v3f[1]*p, v3f[2]*p, seed);
            fsum += f;
            seed += 7;
        }
        sum /= fsum;
        c[j] = sig(12*sum-1);
    }

    //float z = 100;
    //c[3] = pow(z, 1-(nextrand(nextrand(nextrand(c[3]*987654321))) % 1723) * 0.00058f) / z;

    float p2 = 10001;
    float diffusion2 = 0.5f, shift2 = 0.5f, e2 = 2.0f;
    //float spots1 = sig( -voronoi3(v3f[0]*p1, v3f[1]*p1, v3f[2]*p1, diffusion1, shift1, e1) * 18 - 15 );
    float spots2 = pow( -(0.5f-0.5f*voronoi3(v3f[0]*p2, v3f[1]*p2, v3f[2]*p2, diffusion2, shift2, e2)),14);

    float p1 = 2001;
    float diffusion1 = 0.5f, shift1 = 0.5f, e1 = 2.0f;
    //float spots1 = sig( -voronoi3(v3f[0]*p1, v3f[1]*p1, v3f[2]*p1, diffusion1, shift1, e1) * 18 - 15 );
    float spots1 = pow( -(0.5f-0.5f*voronoi3(v3f[0]*p1, v3f[1]*p1, v3f[2]*p1, diffusion1, shift1, e1)),14);

    float p = 201;
    float diffusion = 0.5f, shift = 0.5f, e = 2.0f;
    //float spots = sig( -voronoi3(v3f[0]*p, v3f[1]*p, v3f[2]*p, diffusion, shift, e) * 18 - 15 );
    float spots = pow( -(0.5f-0.5f*voronoi3(v3f[0]*p, v3f[1]*p, v3f[2]*p, diffusion, shift, e)),14);

    float stars = 1.29f;
    float nebula = 0.1f;
    //c[3] = stars*exposure( spots1*0.6 + 0.8*spots );
    c[3] = stars*(spots2*0.8f + spots1*0.9f + 0.99f*spots );
    
    c4f[0] = xmin(1, 0.00f + 0.91f * nebula * c[0] + c[3]);
    c4f[1] = xmin(1, 0.00f + 0.91f * nebula * c[1] + c[3]);
    c4f[2] = xmin(1, 0.05f + 0.99f * nebula * c[2] + c[3]);
    c4f[3] = 1;
}


static float getGround(float* v3f, float* c4f) {
    const float scale = 4*512;
    const int seed1 = 11;
    const float p1 = scale / 101.0f;
    float o1 = simplex3(37 + v3f[0]*p1, 12 + v3f[1]*p1, 97 + v3f[2]*p1, seed1);

    const int seed2 = seed1+43;
    const float p2 = scale / 47.0f;
    float o2 = simplex3(71 + v3f[0]*p2, 45 + v3f[1]*p2, 19 + v3f[2]*p2, seed2);

    float h = (0.9f * o1 + 0.5f * o2) * 0.5 + 0.5;

    const int seed3 = seed1+43;
    const float p3 = scale / 17.0f;
    float o3 = simplex3(21 + v3f[0]*p3, 41 + v3f[1]*p3, 79 + v3f[2]*p3, seed3);

    float top[] = { 0.59f, 0.50f, 0.18f };
    float bottom[] = { 0.75f, 0.73f, 0.28f };
    float splat[] = { 0.43f, 0.35f, 0.21f };
    float top_alpha = h;
    float bottom_alpha = 1.0f - h;
    float splat_alpha = (exp(o2 * 0.5 + 0.5))*0.1 + (exp(o3 * 0.5 + 0.5))*0.05;

    c4f[0] = top[0] * top_alpha + bottom[0] * bottom_alpha + splat[0] * splat_alpha;
    c4f[1] = top[1] * top_alpha + bottom[1] * bottom_alpha + splat[1] * splat_alpha;
    c4f[2] = top[2] * top_alpha + bottom[2] * bottom_alpha + splat[2] * splat_alpha;
    c4f[3] = 1;
    
    return h;
}


/**
 * ~[0,1] normalized lavatic landscape.
 */
static float getLava(float* v3f, float* color3f) {
    const float scale = 128;

    float x = scale*v3f[0];
    float y = scale*v3f[1];
    float z = scale*v3f[2];

    float power = 5;
    float wlen = 1.0f/ (2*power);
    float dx = power*simplex3(37 + x*wlen, 56 + y*wlen, 97 + z*wlen, 241);
    float dz = power*simplex3(67 + x*wlen, 31 + y*wlen, 17 + z*wlen, 21);
    x += dx;
    z += dz;

    const int seed1 = 117;
    // Prime periods => large smallest common multiple.
    const float p1 = 1.0f / 101.0f;

    float x_ = 90 + x * p1;
    float y_ = 39 + y * p1;
    float z_ = 10 + z * p1;

    float o1 = simplex3(x_, y_, z_, seed1);

    const float a = 0.5f * p1;
    const float b = a * 0.5f;
    const float c = b * 0.5f;

    float o1_ = simplex3(x_ + a, y_ + c, z_ + b, seed1);

    // Limit to about [0,1] range.
    float h = 1.0f-(xmin(o1, -o1_)*0.98f+0.96f);
    //h = (h < 0.999999f) ? h : 0.999999f;

    //float diffusion = 0.9;
    //float shift = 0.5;
    //h = 1.0f-voronoi3(x_, 0, z_,diffusion,shift);

    float ash[] = { 0.3f, 0.3f, 0.35f };
    float lava[] = { 1.0f, 0.0f, 0.0f };
    float glow[] = { 1.1f, 1.1f, 1.1f };
    float ash_alpha = h*h;
    float lava_alpha = 0.25f*(2.0f-2.0f*h)*(2.0f-2.0f*h);
    float glow_alpha = (h < 0.2f) ? 1.2f-5.0f*h : 0.0f;
    color3f[0] = sig(6*(lava[0] * lava_alpha + ash[0] * ash_alpha + glow[0] * glow_alpha)-3);
    color3f[1] = sig(6*(lava[1] * lava_alpha + ash[1] * ash_alpha + glow[1] * glow_alpha)-3);
    color3f[2] = sig(6*(lava[2] * lava_alpha + ash[2] * ash_alpha + glow[2] * glow_alpha)-3);

    return h;
}


static void getCloud(float* v3f, float* c4f) {

    unsigned char seed = 31;
    float sum = 0;
    float fsum = 0;
    for (int i = 0; i < 8; i++) {
        float p = 1 << i;
        float f = (8 - i)*2;
        sum += f * simplex3(v3f[0]*p, v3f[1]*p, v3f[2]*p, seed);
        fsum += f;
        seed += 7;
    }
    sum /= fsum;

    float f00 = sig(9*sum);

    float covering = 0.55f;
    f00 = xmax(0, f00 - (1.0f - covering)) / covering;

    float sharpness = 0.98f;
    f00 = pow(f00, 2.0f - sharpness);

    float f01 = sin( 2.14 * f00 );

    c4f[0] = f01;
    c4f[1] = f01;
    c4f[2] = f01;
    c4f[3] = f00;
}


static float continentalRidge(float x) {
    float a = pow(x*2.05f, 12);//1.1f * x;
    //float b = 0.5f * x + 0.26f;
    float b = 0.61f * x + 0.197f;
    float c = 0.51f;
    float d = pow(1.0f-fabs(x-0.57), 11);
    float h = xmax(d, xmin(a, xmin(b, c)));
    return h;
}

static float getContinental(float* v3f, float* c4f) {
    float x = v3f[0];
    float y = v3f[1];
    float z = v3f[2];

    float scale = 128;

    int seed = 11;
    const int n = 8;
    float o = 0;
    const float p[] = { 1.0f/5003.0f, 1.0f/1001.0f, 1.0f/301.0f, 1.0f/47.0f, 1.0f/23.0f, 1.0f/11.0f, 1.0f/ 7.0f, 1.0f/3.0f };
    const float sx[] = { 231,123,45,31,13,7,3,1 };
    float factor = 0.9;
    x *= 0.1;
    y *= 0.1;
    z *= 0.1;
    loopi(n) {
        o += factor * simplex3(sx[i] + x*p[i]*scale, sx[i] + y*p[i]*scale, sx[i] + z*p[i]*scale, seed);
        factor *= 0.8;
        seed = (seed+1)<<1;
    }

    float h = continentalRidge(o*0.5f+0.5f);

    float sea[] = { 0.1,0.1,0.3 };
    float beach[] = { 0.7, 0.7, 0.5 };
    //float grass[] = { 0.3, 0.7, 0.3 };
    //float peak[] = { 1,1,1 };
    float peak[] = { 0.4f,0.3f,0 };

    float sea_alpha = (1.0f - h)*(1.0f - h);
    float beach_alpha = (1.0f - xabs(h - 0.5))*(1.0f - xabs(h - 0.5));
    float peak_alpha = h*h;

    loopi(3) {
        c4f[i] = exposure(5*(sea[i] * sea_alpha + beach[i] * beach_alpha + peak[i] * peak_alpha));
    };
    c4f[3] = 1;

    return h;
}


cDomeBackground::cDomeBackground() {

    struct Texture {
        const char* filename;
        bool soft, repeat;
        int id;
    } texturedefs[] = {
        { "data/planet/background/mountain.tga", true, false, T_MOUNTAIN},
    };

    loopi(1) {
        //if (i == 1) continue;
        int w, h, bpp;
        unsigned char* texels = loadTGA(texturedefs[i].filename, &w, &h, &bpp);
        bool soft = texturedefs[i].soft;
        bool repeat = texturedefs[i].repeat;
        unsigned int texname;
        glBindTexture2D(texname, true, soft, repeat, repeat, w, h, bpp, texels);
        int id = texturedefs[i].id;
        textures[id] = texname;
        delete texels;
    }

    // Ground Texture
    if (1) {
        cout << "Generating Ground..." << endl;
        int w = 1 << (8+DETAIL);
        int h = w;
        int bpp = 3;
        unsigned char* texels = new unsigned char[((unsigned long)w)*h*bpp];
        unsigned char* p = texels;
        loopij(h,w) {

            float dx = ((float)j / (float)w);
            float dz = ((float)i / (float)h);
            float v3f[] = { dx, 0, dz };

            float c = sig(voronoi3_16(dx*16, 0, dz*16) * 8 + 4);
            float c4f[4] = { c, c, c };

            *p++ = 255*c4f[2];
            *p++ = 255*c4f[1];
            *p++ = 255*c4f[0];
            //*p++ = 255*c4f[3];
        }
        bool soft = false;
        bool repeat = true;
        unsigned int texname;
        glBindTexture2D(texname, true, soft, repeat, repeat, w, h, bpp, texels);
        textures[T_ICE] = texname;
        delete texels;
    }

    // Sun Texture
    if (1) {
        cout << "Generating Sun..." << endl;
        int w = 1 << (7+DETAIL);
        int h = w;
        int bpp = 4;
        unsigned char* texels = new unsigned char[((unsigned long)w)*h*bpp];
        unsigned char* p = texels;
        loopij(h,w) {

            float dx = (((float)j / (float)w) - 0.5f) * 2.0f;
            float dy = (((float)i / (float)h) - 0.5f) * 2.0f;
            float r = sqrt(dx*dx+dy*dy);
            float r_ = xmin(1, r);
            float a = 1.0f-pow(r_, 0.9);

            const float f = 2.0f;
            float v3f[] = { f * (float)j / (float)w, 0, f * (float)i / (float)h };

            float s = 4;
            float c = 0.5f + 0.5f * simplex3(s*f * (float)j / (float)w, 0, s*f * (float)i / (float)h, 21);

            float c4f[4] = {
                exposure(3*a + 3*c*a),
                exposure(3*a),
                exposure(1*a),
                a*a
            };

            //getCloud(v3f, c4f);
            *p++ = 255*c4f[2];
            *p++ = 255*c4f[1];
            *p++ = 255*c4f[0];
            *p++ = 255*c4f[3];
        }
        bool soft = true;
        bool repeat = true;
        unsigned int texname;
        glBindTexture2D(texname, true, soft, repeat, repeat, w, h, bpp, texels);
        textures[T_SUN] = texname;
        delete texels;
    }

    // Earth Texture
    if (1) {
        cout << "Generating Earth..." << endl;
        int w = 1 << (7+DETAIL);
        int h = w;
        int bpp = 4;
        unsigned char* texels = new unsigned char[((unsigned long)w)*h*bpp];
        unsigned char* p = texels;
        loopij(h,w) {

            float dx = (((float)j / (float)w) - 0.5f) * 2.0f;
            float dy = (((float)i / (float)h) - 0.5f) * 2.0f;
            float r = sqrt(dx*dx+dy*dy);

            float dz = sqrt(1 - dx*dx - dy*dy);

            float r_ = xmin(1, r);
            float a = 1.0f-pow(r_, 16);

            const float f = 2.0f;
            float v3f[] = { f * dx, f * dy, f * dz };

            float c4f[4];
            getCloud(v3f, c4f);

            float c4f_[4];
            getContinental(v3f, c4f_);

            *p++ = 255*xmin(1, +0.2*c4f[2]+ 0.99*c4f_[2])*a;
            *p++ = 255*xmin(1, +0.2*c4f[1]+ 0.99*c4f_[1])*a;
            *p++ = 255*xmin(1, +0.2*c4f[0]+ 0.99*c4f_[0])*a;
            *p++ = 255*a;
        }
        bool soft = true;
        bool repeat = true;
        unsigned int texname;
        glBindTexture2D(texname, true, soft, repeat, repeat, w, h, bpp, texels);
        textures[T_EARTH] = texname;
        delete texels;
    }

    // Lavos Texture
    if (1) {
        cout << "Generating Lavos..." << endl;
        int w = 1 << (7+DETAIL);
        int h = w;
        int bpp = 4;
        unsigned char* texels = new unsigned char[((unsigned long)w)*h*bpp];
        unsigned char* p = texels;
        loopij(h,w) {

            float dx = (((float)j / (float)w) - 0.5f) * 2.0f;
            float dy = (((float)i / (float)h) - 0.5f) * 2.0f;
            float r = sqrt(dx*dx+dy*dy);

            float dz = sqrt(1 - dx*dx - dy*dy);

            float r_ = xmin(1, r);
            float a = 1.0f-pow(r_, 16);

            const float f = 2.0f;
            float v3f[] = { f * dx, f * dy, f * dz };

            //v3f[0] *= 0.1f;
            float c4f[4];
            getCloud(v3f, c4f);
            //v3f[0] *= 10.0;

            float c4f_[4];
            getLava(v3f, c4f_);

            *p++ = 255*xmax(0, -0.2*c4f[2]+ 0.99*c4f_[2])*a;
            *p++ = 255*xmax(0, -0.2*c4f[1]+ 0.99*c4f_[1])*a;
            *p++ = 255*xmax(0, -0.2*c4f[0]+ 0.99*c4f_[0])*a;
            //*p++ = 255*(0.2*c4f[2] + 0.8*c4f_[2]);
            //*p++ = 255*(0.2*c4f[1] + 0.8*c4f_[1]);
            //*p++ = 255*(0.2*c4f[0] + 0.8*c4f_[0]);
            *p++ = 255*a;
        }
        bool soft = true;
        bool repeat = true;
        unsigned int texname;
        glBindTexture2D(texname, true, soft, repeat, repeat, w, h, bpp, texels);
        textures[T_LAVOS] = texname;
        delete texels;
    }

    // Moon Texture
    if (1) {
        cout << "Generating Moon..." << endl;
        int w = 1 << (7+DETAIL);
        int h = w;
        int bpp = 4;
        unsigned char* texels = new unsigned char[((unsigned long)w)*h*bpp];
        unsigned char* p = texels;
        loopij(h,w) {

            float dx = (((float)j / (float)w) - 0.5f) * 2.0f;
            float dy = (((float)i / (float)h) - 0.5f) * 2.0f;
            float r = sqrt(dx*dx+dy*dy);

            float dz = sqrt(1 - dx*dx - dy*dy);

            float r_ = xmin(1, r);
            float a = 1.0f-pow(r_, 16);

            const float f = 2.0f;
            float v3f[] = { f * dx, f * dy, f * dz };

            v3f[0] *= 0.1f;
            float c4f[4];
            getCloud(v3f, c4f);
            v3f[0] *= 10.0;

            float c4f_[4];
            getGround(v3f, c4f_);

            *p++ = 255*xmax(0.8*c4f[2], 0.8*c4f_[2]);
            *p++ = 255*xmax(0.8*c4f[1], 0.8*c4f_[1]);
            *p++ = 255*xmax(0.8*c4f[0], 0.8*c4f_[0]);
            //*p++ = 255*(0.2*c4f[2] + 0.8*c4f_[2]);
            //*p++ = 255*(0.2*c4f[1] + 0.8*c4f_[1]);
            //*p++ = 255*(0.2*c4f[0] + 0.8*c4f_[0]);
            *p++ = 255*a;
        }
        bool soft = true;
        bool repeat = true;
        unsigned int texname;
        glBindTexture2D(texname, true, soft, repeat, repeat, w, h, bpp, texels);
        textures[T_MOON] = texname;
        delete texels;
    }

    // Cloud Texture
    if (1) {
        cout << "Generating Clouds..." << endl;
        int w = 1 << (9+DETAIL);
        int h = w;
        int bpp = 4;
        unsigned char* texels = new unsigned char[((unsigned long)w)*h*bpp];
        unsigned char* p = texels;
        loopij(h,w) {

            float dx = (((float)j / (float)w) - 0.5f) * 2.0f;
            float dy = (((float)i / (float)h) - 0.5f) * 2.0f;
            float r = sqrt(dx*dx+dy*dy);
            float r_ = xmin(1, r);
            float a = 1.0f-pow(r_, 1.2);

            const float f = 2.0f;
            float v3f[] = { f * (float)j / (float)w, 0, f * (float)i / (float)h };
            float c4f[4];
            getCloud(v3f, c4f);
            *p++ = 255*c4f[2];
            *p++ = 255*c4f[1];
            *p++ = 255*c4f[0];
            *p++ = 255*c4f[3]*a;
        }
        bool soft = true;
        bool repeat = true;
        unsigned int texname;
        glBindTexture2D(texname, true, soft, repeat, repeat, w, h, bpp, texels);
        textures[T_CLOUDS] = texname;
        delete texels;
    }

    // Galaxy Box Textures
    if (1) {
        cout << "Generating Galaxy..." << endl;
        int w = 1 << (8+DETAIL);
        int h = w;
        int bpp = 3;
        unsigned char* texels = new unsigned char[((unsigned long)w)*h*bpp];
        loopk(6) {
            // k even => -1 neg, k odd => +1 pos
            float sign = -1 + 2 * (k & 1);
            // rotation = k / 2
            int rotation = k >> 1;

            cout << "Generating Galaxy for " << ((sign<0)?"-":"+") << " Axis " << rotation << "..." << endl;

            unsigned char* p = texels;
            loopij(h,w) {
                const float f = 0.1;
                const float t = f;

                // Project onto Sphere
                float u = 2.0f * ((float)j / (float)(w)) - 1.0f;
                float v = 2.0f * ((float)i / (float)(h)) - 1.0f;
                float a = sign;
                float oolen = 1.0f / sqrt(u*u + v*v + 1.0f);
                u *= oolen;
                v *= oolen;
                a *= oolen;

                float v3f[] = { a, u, v };
                loop(l, rotation) {
                    float tmp = v3f[2];
                    v3f[2] = v3f[1];
                    v3f[1] = v3f[0];
                    v3f[0] = tmp;
                }

                v3f[0] = v3f[0] * f + t;
                v3f[1] = v3f[1] * f + t;
                v3f[2] = v3f[2] * f + t;

                float c4f[4];
                getNebula(v3f, c4f);
                *p++ = 255*c4f[2];
                *p++ = 255*c4f[1];
                *p++ = 255*c4f[0];
                //*p++ = 255*c4f[3];
            }
            const bool soft = true;
            const bool repeat = false;
            unsigned int texname;
            glBindTexture2D(texname, true, soft, repeat, repeat, w, h, bpp, texels);
            textures[T_GALAXY_NEGX + k] = texname;
        }
        delete texels;
    }


    heightshift = -10;
}

void cDomeBackground::drawBackground(float hour) {
    // Push random seed because the current seed is going to be
    // replaced by a (deterministic) constant value for clouds.
    unsigned int seed = rand();

    //float speed = 10000;
    float speed = 1;
    this->hour = fmod(speed*hour,24.00f);
    //hour = rand()%24;

    // hour = [0,24]
    // => [-90,   0, 90, 180, 270]
    // => [ -1,   0,  1,   0,  -1] = sin
    // => [  0, 0.5,  1, 0.5,   0] = s brightness sine-wave.
    float t = ((this->hour * 360.0 / 24.0) - 90.0); // = [-90,+270]
    float s = sin(t * 0.017453)*0.5 + 0.5; // = [0,1]

    // Set gradient colors according to brightness sine-wave.
    vector_set(topColor, 0.0, 0.0, s * 0.8 + 0.2); // original
    vector_set(middleColor, s * 0.95 + 0.05, s * 0.95 + 0.05, s * 0.95 + 0.05); // original
    //vector_set(bottomColor, 0.4 * s + 0.1, 0.2 * s + 0.1, 0.1 * s + 0.1); // original
    vector_set(bottomColor, s * 0.95 + 0.05, s * 0.95 + 0.05, s * 0.95 + 0.05);

    //vector_set(mBtColor, 0.1 * s + 0.0, 0.1 * s + 0.0, 0.1 * s + 0.0);
    //vector_set(mBtColor, 0.4 * s + 0.1, 0.4 * s + 0.1, 0.4 * s + 0.1);
    //vector_set(mBtColor, 0.6 * s + 0.1, 0.6 * s + 0.1, 0.9 * s + 0.1);
    //vector_set(mBtColor, 0.44 * s + 0.1, 0.44 * s + 0.1, 0.46 * s + 0.1);
    //vector_set(mBtColor, 0.24 * s + 0.1, 0.44 * s + 0.1, 0.26 * s + 0.1);

    if (1) {
        float p[] = {700, 100, 0, 1};
        float a[] = {0.3, 0.3, 0.3, 1};
        glLightfv(GL_LIGHT0, GL_POSITION, p);
        glLightfv(GL_LIGHT0, GL_AMBIENT, a);
        glLightf(GL_LIGHT0, GL_CONSTANT_ATTENUATION, 1.0);
        glLightf(GL_LIGHT0, GL_LINEAR_ATTENUATION, 0.0001);
        glLightf(GL_LIGHT0, GL_QUADRATIC_ATTENUATION, 0.000001);
        glEnable(GL_NORMALIZE);
    }


    GLint mode;
    glGetIntegerv(GL_POLYGON_MODE, &mode);
    if (mode == GL_FILL) {
        drawGalaxy();
        drawOrbit();
        drawUpperDome();
        drawSun();
        drawClouds();
        //drawMountains();
        drawLowerDome();
        //drawGround();
        if (!true) drawRain();
    } else {
        drawOrbit();
        //drawUpperDome();
        //drawMountains();
        //drawLowerDome();
    }

    float ca[4] = {
        bottomColor[0],
        bottomColor[1],
        bottomColor[2],
        0.0f
    };
    glFogfv(GL_FOG_COLOR, ca);

    // Pop a seed.
    srand(seed);
}

void cDomeBackground::drawGalaxy() {
    glPushAttrib(GL_ENABLE_BIT | GL_CURRENT_BIT | GL_TEXTURE_BIT);
    {
        glDisable(GL_DEPTH_TEST);
        glDisable(GL_LIGHTING);
        glEnable(GL_TEXTURE_2D);
        glDisable(GL_FOG);
        glEnable(GL_CULL_FACE);

        glColor4f(1,1,1,1);

        glPushMatrix();
        {
            // Load current Camera Matrix and
            // set Position (4th Column) = 0.
            float m[16];
            glGetFloatv(GL_MODELVIEW_MATRIX, m);
            loop3i(m[12 + i] = 0);
            glLoadIdentity();
            glMultMatrixf(m);
            glScalef(100, 100, 100);

            float angle = 180 - ((360 / 24.00f) * hour + 90);
            // Sun travels: dawn 0600:East, 1200:South, dusk 1800:West (2400:North -> other side of the globe).
            // 0000/2400: 0 degree, 0600: 90 degree, 1200: 180 degree, 1800: 270 degree, ...
            // Moon is traveling about 180 degree behind/ahead.
            // Well, that is, if you're down-under then east, north, west (not east, south, west).
            glRotatef(-90, 0, 1, 0);
            //	glRotatef(45+7.5, 0, 0, 1); // Longitude? Aquator Angle Offset -> Germany. and for example Poland, GB and Canada.
            glRotatef(55, 0, 0, 1);
            glRotatef(0, 0, 0, 1); // Aquator.

            glRotatef(180+angle,1,0,0);

            // NEG X
            if (1) {
            glBindTexture(GL_TEXTURE_2D, textures[T_GALAXY_NEGX]);
            glBegin(GL_QUADS);
                glTexCoord2i(0, 0);
                glVertex3f(-1, -1, -1);
                glTexCoord2i(1, 0);
                glVertex3f(-1, +1, -1);
                glTexCoord2i(1, 1);
                glVertex3f(-1, +1, +1);
                glTexCoord2i(0, 1);
                glVertex3f(-1, -1, +1);
            glEnd();
            }
            // POS X
            glBindTexture(GL_TEXTURE_2D, textures[T_GALAXY_POSX]);
            glBegin(GL_QUADS);
                glTexCoord2i(1, 0);
                glVertex3f(+1, +1, -1);
                glTexCoord2i(0, 0);
                glVertex3f(+1, -1, -1);
                glTexCoord2i(0, 1);
                glVertex3f(+1, -1, +1);
                glTexCoord2i(1, 1);
                glVertex3f(+1, +1, +1);
            glEnd();
            // NEG Y
            if (1) {
            glBindTexture(GL_TEXTURE_2D, textures[T_GALAXY_NEGY]);
            glBegin(GL_QUADS);
                glTexCoord2i(0, 0);
                glVertex3f(-1, -1, -1);
                glTexCoord2i(1, 0);
                glVertex3f(-1, -1, +1);
                glTexCoord2i(1, 1);
                glVertex3f(+1, -1, +1);
                glTexCoord2i(0, 1);
                glVertex3f(+1, -1, -1);
            glEnd();
            }
            // POS Y
            glBindTexture(GL_TEXTURE_2D, textures[T_GALAXY_POSY]);
            glBegin(GL_QUADS);
                glTexCoord2i(1, 0);
                glVertex3f(-1, +1, +1);
                glTexCoord2i(0, 0);
                glVertex3f(-1, +1, -1);
                glTexCoord2i(0, 1);
                glVertex3f(+1, +1, -1);
                glTexCoord2i(1, 1);
                glVertex3f(+1, +1, +1);
            glEnd();
            // NEG Z
            if (1) {
            glBindTexture(GL_TEXTURE_2D, textures[T_GALAXY_NEGZ]);
            glBegin(GL_QUADS);
                glTexCoord2i(0, 0);
                glVertex3f(-1, -1, -1);
                glTexCoord2i(1, 0);
                glVertex3f(+1, -1, -1);
                glTexCoord2i(1, 1);
                glVertex3f(+1, +1, -1);
                glTexCoord2i(0, 1);
                glVertex3f(-1, +1, -1);
            glEnd();
            }
            // POS Z
            glBindTexture(GL_TEXTURE_2D, textures[T_GALAXY_POSZ]);
            glBegin(GL_QUADS);
                glTexCoord2i(1, 0);
                glVertex3f(+1, -1, +1);
                glTexCoord2i(0, 0);
                glVertex3f(-1, -1, +1);
                glTexCoord2i(0, 1);
                glVertex3f(-1, +1, +1);
                glTexCoord2i(1, 1);
                glVertex3f(+1, +1, +1);
            glEnd();
        }
        glPopMatrix();
    }
    glPopAttrib();
}


void cDomeBackground::drawUpperDome() {
    glPushAttrib(GL_ALL_ATTRIB_BITS | GL_ENABLE_BIT | GL_CURRENT_BIT);
    {
        glDisable(GL_DEPTH_TEST);
        glDisable(GL_LIGHTING);
        glEnable(GL_CULL_FACE);
        glDisable(GL_FOG);
        //glBlendFunc( GL_ONE_MINUS_SRC_ALPHA, GL_SRC_ALPHA );
        glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );

        // Load current Camera Matrix and
        // set Position (4th Column) = 0.
        float m[16];
        glGetFloatv(GL_MODELVIEW_MATRIX, m);
        loop3i(m[12 + i] = 0);
        m[12 + 1] = heightshift;

        float light = 0.1 + 0.8*cos((hour-12.0f)/12.00f*0.5f*M_PI);
        float density = light;

        glPushMatrix();
        {
            glLoadIdentity();
            glMultMatrixf(m);
            //glScalef(100, 100 * 0.3, 100); //*0.25 original
            glScalef(100, 100 * 0.25, 100); // original

            int vsteps = 8;
            double vf = 0.5 * M_PI / (double) vsteps;

            int hsteps = 18;
            double hf = 2.0 * M_PI / (double) hsteps;
            double h = 0;

            float color1[3];
            vector_cpy(color1, topColor);

            float color2[3];
            vector_cpy(color2, middleColor);

            float slope[3];
            vector_set3i(slope, (color2[i] - color1[i]) / (float) vsteps);

            vector_cpy(color2, color1);
            vector_add(color2, color2, slope);
            //glColor3f(0.25,0.25,0.25);
            float s1 = sin(0);
            float c1 = cos(0);
            float density1 = (1 - c1) + c1 * density;
            for (int i = 1; i <= vsteps; i++) {
                float s2 = sin(i * vf);
                float c2 = cos(i * vf);
                float density2 = (1 - c2) + c2 * density;
                glBegin(GL_TRIANGLE_STRIP);
                for (int j = 0; j <= hsteps; j++) {
                    float s = sin(h);
                    float c = cos(h);
                    glColor4f(color1[0],color1[1],color1[2],density1);
                    //glTexCoord2f(0.5f+0.5f*s1*c, 0.5f+0.5f*s1*s);
                    glNormal3f(s1*c, c1, s1 * s);
                    glVertex3f(s1*c, c1, s1 * s);
                    glColor4f(color2[0],color2[1],color2[2],density2);
                    //glTexCoord2f(0.5f+0.5f*s2*c, 0.5f+0.5f*s2*s);
                    glNormal3f(s2*c, c2, s2 * s);
                    glVertex3f(s2*c, c2, s2 * s);
                    h += hf;
                } // for hsteps
                glEnd();
                vector_cpy(color1, color2);
                vector_add(color2, color2, slope);
                s1 = s2;
                c1 = c2;
                density1 = density2;
            } // for vsteps
        }
        glPopMatrix();
    }
    glPopAttrib();
}

void cDomeBackground::drawLowerDome() {
    glPushAttrib(GL_ENABLE_BIT | GL_CURRENT_BIT);
    {
        glDisable(GL_DEPTH_TEST);
        glDisable(GL_LIGHTING);
        glEnable(GL_CULL_FACE);
        glDisable(GL_FOG);

        // Load current Camera Matrix and
        // set Position (4th Column) = 0.
        float m[16];
        glGetFloatv(GL_MODELVIEW_MATRIX, m);
        loop3i(m[12 + i] = 0);
        m[12 + 1] = heightshift;

        glPushMatrix();
        {
            glLoadIdentity();
            glMultMatrixf(m);
            glScalef(100, 100 * 0.1, 100);

            int vsteps = 8;
            double vf = 0.5f * M_PI / (double) vsteps;

            int hsteps = 18;
            double hf = 2.0 * M_PI / (double) hsteps;
            double h = 0;

            const float f = 0.75;
            float color2[3] = {0.4, 0.2, 0.1};
            vector_cpy(color2, bottomColor);
            float color1[3] = {color2[0] * f, color2[1] * f, color2[2] * f};

            float tmp[3];
            vector_cpy(tmp, color1);
            vector_cpy(color1, color2);
            vector_cpy(color2, tmp);

            float slope[3];
            vector_set3i(slope, (color2[i] - color1[i]) / (float) vsteps);

            vector_cpy(color2, color1);
            vector_add(color2, color2, slope);

            float s1 = -sin(0);
            float c1 = -cos(0);
            for (int i = 1; i <= vsteps; i++) {
                float s2 = -sin(i * vf);
                float c2 = -cos(i * vf);
                glBegin(GL_TRIANGLE_STRIP);
                for (int j = 0; j <= hsteps; j++) {
                    float s = sin(h);
                    float c = cos(h);
                    glColor3fv(color2);
                    glNormal3f(s2*c, c2, s2 * s);
                    glVertex3f(s2*c, c2, s2 * s);
                    glColor3fv(color1);
                    glNormal3f(s1*c, c1, s1 * s);
                    glVertex3f(s1*c, c1, s1 * s);
                    h += hf;
                };
                glEnd();
                vector_cpy(color1, color2);
                vector_add(color2, color2, slope);
                s1 = s2;
                c1 = c2;
            } // for vsteps
        }
        glPopMatrix();
    }
    glPopAttrib();
}

void cDomeBackground::drawGround() {
    //return;
    glPushAttrib(GL_ENABLE_BIT | GL_CURRENT_BIT | GL_TEXTURE_BIT);
    {
        glDisable(GL_LIGHTING);
        glEnable(GL_TEXTURE_2D);
        glDisable(GL_FOG);
        glDisable(GL_CULL_FACE);
        glBindTexture(GL_TEXTURE_2D, textures[T_ICE]);

        // Get current Camera-Matrix.
        float m[16];
        glGetFloatv(GL_MODELVIEW_MATRIX, m);
        // Position from 4th Col.
        float p[3];
        loop3i(p[i] = m[12 + i]);
        // Reset Position on 4th Col.
        loop3i(m[12 + i] = 0);
        // Apply transpose-inverse to position.
        float n[16];
        memcpy(n, m, sizeof (float) *16);
        matrix_transpose(m);
        matrix_apply2(m, p);
        //cout << "height over ground: " << p[1] << "\n";
        //cout << "Position: " << p[0] << " " << p[2] << "\n";

        glPushMatrix();
        {
            glLoadIdentity();
            glMultMatrixf(n); // Transpose to undo previous transpose.
            glTranslatef(0, p[Y], 0);

            float color2[3] = {0.4, 0.2, 0.1};
            vector_cpy(color2, bottomColor);

            //const float tdiv = 5; // original
            const float tdiv = 16;
            const float radius = 400;

            float dx = -p[0] / tdiv;
            float dz = -p[2] / tdiv;

            glBegin(GL_TRIANGLE_FAN);
            glColor4f(bottomColor[0], bottomColor[1], bottomColor[2], 1.0f);
//glColor4f(0.1,0.1,0.3,0.93); // water
glColor4f(1,1,1,1);
            glNormal3f(0, 1, 0);
            glTexCoord2f(dx, dz);
            glVertex3f(0.0, 0.0, 0.0);
            glColor4f(bottomColor[0], bottomColor[1], bottomColor[2], 0.0f);
//glColor4f(0.1,0.1,0.4,0); // water
glColor4f(1,1,1,1);
            float e = 20.0 * M_PI / 180.0f;

            loopi(19) {
                glNormal3f(0, 1, 0);
                glTexCoord3f(cos(-i * e) * radius / tdiv + dx, sin(-i * e) * radius / tdiv + dz, 0);
                glVertex3f(cos(-i * e) * radius, 0, sin(-i * e) * radius);
            } // loopi 19
            glEnd();
        }
        glPopMatrix();
    }
    glPopAttrib();
}

void cDomeBackground::drawClouds() {
    glPushAttrib(GL_ENABLE_BIT | GL_CURRENT_BIT);
    {
        glDisable(GL_DEPTH_TEST);
        glDisable(GL_LIGHTING);
        glDisable(GL_CULL_FACE);
        glDisable(GL_FOG);
        glEnable(GL_TEXTURE_2D);

        glBindTexture(GL_TEXTURE_2D, textures[T_CLOUDS]);

        // Load current Camera Matrix and
        // set Position (4th Column) = 0.
        float m[16];
        glGetFloatv(GL_MODELVIEW_MATRIX, m);
        loop3i(m[12 + i] = 0);
        //m[12 + 0] = m[12 + 2] = 0;
        m[12 + 1] = heightshift;

        glPushMatrix();
        {
            glLoadIdentity();
            glMultMatrixf(m);
            glScalef(100, 100 * 0.24, 100);
            //glScalef(100, 100 * 0.125, 100); // original

            float windspeed = 2;
            float trad = windspeed * 24.0f * (hour / 24.00f) * 2 * M_PI;
            float light = 0.1 + 0.8*cos((hour-12.0f)/12.00f*0.5f*M_PI);
            glRotatef(trad / 0.017453f, 1, 0.1, 1);

            //cout << hour << endl;
            srand(12421);
            int n = 5;
            float density = 3.0f / float(1.0f + n);
            density = 0.99;
            for (int i = 0; i < n; i++) {
                //float xrad = -(rand() % 157) * 0.01f;
                float xrad = -(157 - (rand() % 314)) * 0.01f;
                float yrad = (rand() % 628) * 0.01f;

                /*
                float xaxis[] = { 1,0,0 };
                float qx[4];
                quat_rotaxis(qx, xrad, xaxis);

                float yaxis[] = { 0,1,0 };
                float qy[4];
                quat_rotaxis(qy, yrad, yaxis);

                float qxy[4];
                quat_mul(qxy, qx, qy);

                float x[] = {1,0,0};
                float y[] = {0,1,0};
                float z[] = {0,0,1};
                quat_apply(x, qxy, x);
                quat_apply(y, qxy, y);
                quat_apply(z, qxy, z);

                glColor4f(1,1,1,1);
                glVertex3f(1*x[0]+z[0], 1+z[1], 0+z[2]);
                glVertex3f(0, 1, 0);
                glVertex3f(0, 0, 0);
                glVertex3f(1, 0, 0);
                glEnd();
                 */

                glPushMatrix();
                {
                    glRotatef(yrad / 0.017454, 0, 1, 0);
                    glRotatef(xrad / 0.017454, 1, 0, 0);
                    glTranslatef(0, 0, 1.7);
                    glColor4f(light, light, light, density);
                    glBegin(GL_QUADS);
                    glTexCoord2f(1, 1);
                    glVertex3f(+1, +1, 0);
                    glTexCoord2f(0, 1);
                    glVertex3f(-1, +1, 0);
                    glTexCoord2f(0, 0);
                    glVertex3f(-1, -1, 0);
                    glTexCoord2f(1, 0);
                    glVertex3f(+1, -1, 0);
                    glEnd();
                }
                glPopMatrix();
            }

        }
        glPopMatrix();
    }
    glPopAttrib();
}

void cDomeBackground::drawMountains() {
    srand(762381);
    glPushAttrib(GL_ENABLE_BIT | GL_CURRENT_BIT);
    {
        glEnable(GL_FOG);
        glDisable(GL_DEPTH_TEST);
        glDisable(GL_LIGHTING);
        glDisable(GL_CULL_FACE);
        glEnable(GL_TEXTURE_2D);

        glBindTexture(GL_TEXTURE_2D, textures[T_MOUNTAIN]);
        glColor3fv(bottomColor);

        // Load current Camera Matrix and
        // set Position (4th Column) = 0.
        float m[16];
        glGetFloatv(GL_MODELVIEW_MATRIX, m);
        loop3i(m[12 + i] = 0);
        //m[12 + 0] = m[12 + 2] = 0;
        m[12 + 1] = heightshift;

        glPushMatrix();
        {
            int n = 30;
            float f = 1.0;
            for (int i = 0; i < n; i++) {
                glLoadIdentity();
                glMultMatrixf(m);
                glScalef(90, 100 * 0.2, 90);

                float w = (0.3 + (rand() % 100) * 0.005f) * 0.5 * f;
                float rad = 0.01f * (rand() % 628);
                glRotatef(rad / 0.017453, 0, 1, 0);
                glTranslatef(0, 0, 1);

                float u = 1.2 * f;
                glBegin(GL_QUADS);
                glTexCoord2f(0, 1);
                glVertex3f(-w, u, 1);
                glTexCoord2f(0, 0);
                glVertex3f(-w, -0.01, 1);
                glTexCoord2f(1, 0);
                glVertex3f(+w, -0.01, 1);
                glTexCoord2f(1, 1);
                glVertex3f(+w, u, 1);
                glEnd();

                f = f * 0.95f;
            }
        }
        glPopMatrix();
    }
    glPopAttrib();
}


#include "psi3d/noise.h"

static float getLand_(float x, float y, float z, float* color3f) {
    const int seed1 = 11;
    const float p1 = 1.0f / 101.0f;
    float o1 = simplex3(37 + x*p1, 15 + y*p1, 97 + z*p1, seed1);

    const int seed2 = seed1+43;
    const float p2 = 1.0f / 47.0f;
    float o2 = simplex3(71 + x*p2, 111 + y*p2, 19 + z*p2, seed2);

    float h = sig ((0.9f * o1 + 0.5f * o2) * 1.0f);

    //const int seed3 = seed1+43;
    //const float p3 = 1.0f / 17.0f;
    //float o3 = simplex3(21 + (x+sin(z))*p3, 53 + (y+sin(y))*p3, 79 + (z+sin(x))*p3, seed3);

    float top[] = { 0.59f, 0.80f, 0.58f };
    float bottom[] = { 0.85f, 0.93f, 0.58f };
    float splat[] = { 0.93f, 0.95f, 0.61f };
    float top_alpha = h;
    float bottom_alpha = 1.0f - h;
    float splat_alpha = 0;//(exp(o2 * 0.5 + 0.5))*0.1 + (exp(o3 * 0.5 + 0.5))*0.05;

    color3f[0] = top[0] * top_alpha + bottom[0] * bottom_alpha + splat[0] * splat_alpha;
    color3f[1] = top[1] * top_alpha + bottom[1] * bottom_alpha + splat[1] * splat_alpha;
    color3f[2] = top[2] * top_alpha + bottom[2] * bottom_alpha + splat[2] * splat_alpha;

    return h;
}

inline void transformV3(float* v3f, float* c3f) {
    const float s = 255;
    //float h = 0.1 * sig(simplex3(s*(v[0]+1), s*(v[1]+1), s*(v[2]+1), 83));
    float h = 0.3 * getLand_(s*v3f[0],s*v3f[1],s*v3f[2],c3f);

    v3f[0] = v3f[0] * (1+h);
    v3f[1] = v3f[1] * (1+h);
    v3f[2] = v3f[2] * (1+h);
}

#if 0
inline float InvSqrt (float x){
    float xhalf = 0.5f*x;
    int i = *(int*)&x;
    i = 0x5f3759df - (i>>1);
    x = *(float*)&i;
    x = x*(1.5f - xhalf*x*x);
    return x;
}
#endif

void drawBalloid() {
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_CULL_FACE);
    glDisable(GL_LIGHTING);
    glDisable(GL_TEXTURE_2D);

    int steps = 2;
    double delta = 2.0f / (float) steps;

    float c[3];

    // xy
    loopk (2) {
        float dir = (k & 1) ? -1 : +1;
        glColor4f(0.9f,0.5f,0.5f,1.0f);
        //glColor4f(t==0?1:0.5,t==1?1:0.5,t==2?1:0.5,1);

        double b = -1;
        loopj(steps) {
            double b0 = b * -dir;
            double b1 = (b + delta) * -dir;
            double a = -1;
            glBegin(GL_TRIANGLE_STRIP);
            loopi(steps+1) {
                

                /*
                float alpha = a * M_PI * 0.25f;
                float beta0 = b0 * M_PI * 0.25f;
                float beta1 = b1 * M_PI * 0.25f;
                
                float v0[] = { sin(alpha), 0, cos(alpha) };
                float v1[] = { sin(alpha), 0, cos(alpha) };

                float v0_[3];
                v0_[0] = v0[0];
                v0_[1] = v0[1]*cos(beta0) - v0[2]*sin(beta0);
                v0_[2] = v0[1]*sin(beta0) + v0[2]*cos(beta0);
                vector_cpy(v0, v0_);

                float v1_[3];
                v1_[0] = v1[0];
                v1_[1] = v1[1]*cos(beta1) - v1[2]*sin(beta1);
                v1_[2] = v1[1]*sin(beta1) + v1[2]*cos(beta1);
                vector_cpy(v1, v1_);
                */

                
                float inv0 = 1.0f / (float) sqrt(a*a+b0*b0+1.0f);
                float inv1 = 1.0f / (float) sqrt(a*a+b1*b1+1.0f);
                float v0[] = { a*inv0, b0*inv0, dir*inv0 };
                float v1[] = { a*inv1, b1*inv1, dir*inv1 };
                

                glNormal3fv(v0);
                transformV3(v0,c);
                //glColor3fv(c);
                glVertex3fv(v0);

                glNormal3fv(v1);
                transformV3(v1,c);
                //glColor3fv(c);
                glVertex3fv(v1);

                a += delta;
            }
            glEnd();
            b += delta;
        }
    }

    // xz
    loopk (2) {
        float dir = (k & 1) ? -1 : +1;
        glColor4f(0.5f,0.9f,0.5f,1.0f);
        //glColor4f(t==0?1:0.5,t==1?1:0.5,t==2?1:0.5,1);

        double b = -1;
        loopj(steps) {
            double b0 = b * dir;
            double b1 = (b + delta) * dir;
            double a = -1;
            glBegin(GL_TRIANGLE_STRIP);
            loopi(steps+1) {

                float inv0 = 1.0f / (float) sqrt(a*a+b0*b0+1.0f);
                float inv1 = 1.0f / (float) sqrt(a*a+b1*b1+1.0f);
                float v0[] = { a*inv0, dir*inv0, b0*inv0 };
                float v1[] = { a*inv1, dir*inv1, b1*inv1 };

                glNormal3fv(v0);
                transformV3(v0,c);
                //glColor3fv(c);
                glVertex3fv(v0);

                glNormal3fv(v1);
                transformV3(v1,c);
                //glColor3fv(c);
                glVertex3fv(v1);

                a += delta;
            }
            glEnd();
            b += delta;
        }
    }

    // yz
    loopk (2) {
        float dir = (k & 1) ? -1 : +1;
        glColor4f(0.5f,0.5f,0.9f,1.0f);
        //glColor4f(t==0?1:0.5,t==1?1:0.5,t==2?1:0.5,1);

        double b = -1;
        loopj(steps) {
            double b0 = b * -dir;
            double b1 = (b + delta) * -dir;
            double a = -1;
            glBegin(GL_TRIANGLE_STRIP);
            loopi(steps+1) {

                //a = fma(a,1,0);
                float inv0 = 1.0f / (float) sqrt(a*a+b0*b0+1.0f);
                float inv1 = 1.0f / (float) sqrt(a*a+b1*b1+1.0f);
                float v0[] = { dir*inv0, a*inv0, b0*inv0 };
                float v1[] = { dir*inv1, a*inv1, b1*inv1 };

                glNormal3fv(v0);
                transformV3(v0,c);
                //glColor3fv(c);
                glVertex3fv(v0);
  
                glNormal3fv(v1);
                transformV3(v1,c);
                //glColor3fv(c);
                glVertex3fv(v1);

                a += delta;
            }
            glEnd();
            b += delta;
        }
    }
}


void cDomeBackground::drawSun() {
    float m[16];
    glGetFloatv(GL_MODELVIEW_MATRIX, m);

    // 4th Row = 0.
    m[0 * 4 + 3] = m[1 * 4 + 3] = m[2 * 4 + 3] = 0.0f;

    // 4th Col = 0.
    m[3 * 4 + 0] = m[3 * 4 + 1] = m[3 * 4 + 2] = 0.0f;
    m[3 * 4 + 3] = 1.0f;

    //m[12 + 1] += heightshift;

    glPushAttrib(GL_ALL_ATTRIB_BITS | GL_ENABLE_BIT | GL_CURRENT_BIT);
    {
        //glDisable(GL_CULL_FACE);
        glDisable(GL_DEPTH_TEST);
        glDisable(GL_FOG);
        glDisable(GL_LIGHTING);

        glBlendFunc(GL_SRC_ALPHA, GL_ONE);

        glPushMatrix();
        {
            glLoadIdentity();
            glMultMatrixf(m);
            //glScalef(1,0.3f,1);

            const float scale = 0.1 * 2400.0f;
            float sun = 15 * 80 * 0.1;
            //const float corona = 1.2f;

            static float angle = 0;
            angle -= 1;

            angle = 180 - ((360 / 24.00f) * hour + 90);
            //  angle = 180-((360 / 24.00f) * 20 + 90);

            //angle = 0;

            // Sun travels: dawn 0600:East, 1200:South, dusk 1800:West (2400:North -> other side of the globe).
            // 0000/2400: 0 degree, 0600: 90 degree, 1200: 180 degree, 1800: 270 degree, ...

            // Moon is traveling about 180 degree behind/ahead.

            // Well, that is, if you're down-under then east, north, west (not east, south, west).

            glRotatef(-90, 0, 1, 0);
            //	glRotatef(45+7.5, 0, 0, 1); // Longitude? Aquator Angle Offset -> Germany. and for example Poland, GB and Canada.
            glRotatef(55, 0, 0, 1);
            glRotatef(0, 0, 0, 1); // Aquator.

            const int detail = 7;

            if (!true) { // atmospheric scattering
                glPushMatrix();
                {
                    glRotatef(angle, 1, 0, 0);
                    float dist = 0.005;
                    glBegin(GL_TRIANGLE_FAN);
                    {
                        glColor4f(1, 1, 1, 0.99);
                        glVertex3f(0, 0, scale);
                        float alpha = 0;
                        const float inc = 2.0f * M_PI / float(detail - 1);
                        const float r = scale;

                        loopi(detail) {
                            glColor4f(0, 0, 0.3, 0.1);
                            glVertex3f(sin(alpha) * r, cos(alpha) * r, scale * dist);
                            alpha += inc;
                        }
                    }
                    glEnd();
                    glBegin(GL_TRIANGLE_STRIP);
                    {
                        float alpha = 0;
                        const float inc = 2.0f * M_PI / float(detail - 1);
                        const float r = scale;

                        loopi(detail) {
                            glColor4f(0, 0, 0.3, 0.1);
                            glVertex3f(sin(alpha) * r, cos(alpha) * r, -scale * dist);
                            glColor4f(0, 0, 0.3, 0.1);
                            glVertex3f(sin(alpha) * r, cos(alpha) * r, +scale * dist);
                            alpha -= inc;
                        }
                    }
                    glEnd();
                    glBegin(GL_TRIANGLE_FAN);
                    {
                        glColor4f(1, 1, 1, 0.3);
                        glVertex3f(0, 0, -scale);
                        float alpha = 0;
                        const float inc = 2.0f * M_PI / float(detail - 1);
                        const float r = scale;

                        loopi(detail) {
                            glColor4f(0, 0, 0.3, 0.1);
                            glVertex3f(sin(alpha) * r, cos(alpha) * r, -scale * dist);
                            alpha -= inc;
                        }
                    }
                    glEnd();
                }
                glPopMatrix();
            } // atmospheric scattering

            glPushMatrix();
            //if (hour > 6.50 && hour < 17.50)
            {
                glRotatef(angle, 1, 0, 0);

                glTranslatef(0, 0, scale);
                glColor4f(1, 0.95, 0.9, 0.7);
                //glDisk(detail, sun);
                // Sun
#if 1
                glEnable(GL_TEXTURE_2D);
                glBindTexture(GL_TEXTURE_2D, textures[T_SUN]);
                glBegin(GL_QUADS);
                glTexCoord2f(0, 0);
                glVertex3f(-sun, -sun, 0);
                glTexCoord2f(0, 1);
                glVertex3f(-sun, +sun, 0);
                glTexCoord2f(1, 1);
                glVertex3f(+sun, +sun, 0);
                glTexCoord2f(1, 0);
                glVertex3f(+sun, -sun, 0);
                glEnd();
#else
                glPushMatrix();
                glDisable(GL_CULL_FACE);
                glDisable(GL_TEXTURE_2D);
                sun*=0.5;
                glScalef(sun,sun,sun);
                glRotatef(hour*200000, 1,4,2);
                drawBalloid();
                glPopMatrix();
#endif
            }
            glPopMatrix();

        }
        glPopMatrix();
    }
    glPopAttrib();
}


void cDomeBackground::drawOrbit() {
    float m[16];
    glGetFloatv(GL_MODELVIEW_MATRIX, m);

    // 4th Row = 0.
    m[0 * 4 + 3] = m[1 * 4 + 3] = m[2 * 4 + 3] = 0.0f;

    // 4th Col = 0.
    m[3 * 4 + 0] = m[3 * 4 + 1] = m[3 * 4 + 2] = 0.0f;
    m[3 * 4 + 3] = 1.0f;

    //m[12 + 1] += heightshift;

    glPushAttrib(GL_ENABLE_BIT | GL_CURRENT_BIT);
    {
        //glDisable(GL_CULL_FACE);
        glDisable(GL_DEPTH_TEST);
        glDisable(GL_FOG);
        glDisable(GL_LIGHTING);

        glPushMatrix();
        {
            glLoadIdentity();
            glMultMatrixf(m);
            //glScalef(1,0.3f,1);

            const float scale = 0.1 * 2400.0f;
            float sun = 15 * 80 * 0.1;
            //const float corona = 1.2f;
            float moon = sun * 0.5;

            static float angle = 0;
            angle -= 1;

            angle = 180 - ((360 / 24.00f) * hour + 90);
            //  angle = 180-((360 / 24.00f) * 20 + 90);

            //angle = 0;

            // Sun travels: dawn 0600:East, 1200:South, dusk 1800:West (2400:North -> other side of the globe).
            // 0000/2400: 0 degree, 0600: 90 degree, 1200: 180 degree, 1800: 270 degree, ...

            // Moon is traveling about 180 degree behind/ahead.

            // Well, that is, if you're down-under then east, north, west (not east, south, west).

            glRotatef(-90, 0, 1, 0);
            //	glRotatef(45+7.5, 0, 0, 1); // Longitude? Aquator Angle Offset -> Germany. and for example Poland, GB and Canada.
            glRotatef(55, 0, 0, 1);
            glRotatef(0, 0, 0, 1); // Aquator.

            const int detail = 7;

            glPushMatrix();
            //if (hour > 18.50 || hour < 5.50)
            {
                glRotatef(angle + 180, 1, 0, 0);

                glTranslatef(0, 0, scale);
                //glColor4f(0.6, 0.6, 0.7, 0.99);
                glColor4f(0.8, 0.8, 0.8, 0.99);
                //glDisk(detail, moon);
                // Moon
#if 1
                glEnable(GL_TEXTURE_2D);
                glBindTexture(GL_TEXTURE_2D, textures[T_EARTH]);
                glBegin(GL_QUADS);
                glTexCoord2f(0, 0);
                glVertex3f(-moon, -moon, 0);
                glTexCoord2f(0, 1);
                glVertex3f(-moon, +moon, 0);
                glTexCoord2f(1, 1);
                glVertex3f(+moon, +moon, 0);
                glTexCoord2f(1, 0);
                glVertex3f(+moon, -moon, 0);
                glEnd();
#else
                glPushMatrix();
                glDisable(GL_CULL_FACE);
                glDisable(GL_TEXTURE_2D);
                //moon*=0.5;
                glScalef(moon,moon,moon);
                glRotatef(hour*200000, 1,4,2);
                drawBalloid();
                glPopMatrix();
#endif
            }
            glPopMatrix();

        }
        glPopMatrix();
    }
    glPopAttrib();
}

void cDomeBackground::drawRain() {

    srand(seed);
    seed = rand();
    loopi (rand()%5) {
        if (rain.size() >= 1000) break;
        float dx = 0.3;
        float dz = 0.1;
        cParticle* p = new cParticle();
        float alpha = (rand() % 628) * 0.01;
        float dist = 0.2 + 0.1 * (rand() % 50);
        float h = 15;
        vector_set(p->pos, sin(alpha) * dist - 0.3*dx*h, h, cos(alpha) * dist - 0.3*dz*h);
        vector_cpy(p->old, p->pos);
        vector_set(p->vel, dx, -2 - rand()%3 , dz);
        vector_set(p->fce, 0, 0, 0);
        rain.push_back(p);
    }

    glPushAttrib(GL_ENABLE_BIT | GL_CURRENT_BIT);
    {
        //glDisable(GL_DEPTH_TEST);
        glDisable(GL_LIGHTING);
        glEnable(GL_CULL_FACE);
        //glDisable(GL_FOG);

        glPushMatrix();
        {
            float m[16];
            glGetFloatv(GL_MODELVIEW_MATRIX, m);
            loop3i(m[12 + i] = 0);
            glLoadIdentity();
            glMultMatrixf(m);

            glBegin(GL_LINES);

            foreachNoInc(i, rain) {
                cParticle* p = *i++;
                float end[3];
                vector_scale(end, p->vel, -0.4f);
                vector_add(end, end, p->pos);
                glColor4f(0.8, 0.8, 0.8, 0.3);
                glVertex3fv(end);
                glColor4f(0.3, 0.3, 0.5, 0.9);
                glVertex3fv(p->pos.data());
                p->stepEuler(0.1, 0);
                //std::cout << p->pos[0] << " " << p->pos[1] << " " << p->pos[2] << "\n";
                if (p->pos[1] < -10) rain.remove(p);
            }
            glEnd();
        }
        glPopMatrix();

    }
    glPopAttrib();
}
