/**
 * File:     cBuilding.h
 * Project:  LinWarrior 3D
 * Home:     hackcraft.de
 *
 * Created on March 28, 2008, 10:23 PM
 */

#ifndef BUILDING_H
#define BUILDING_H

class cBuilding;

#include "cObject.h"
#include "cWeapon.h"

/**
 * Models a building of varring size.
 */
class cBuilding : public cObject {
protected:
    static int sInstances;
    static std::map<int,long> sTextures;
protected:
    /// Need to re-setup base matrix for explosion (true/false)?
    bool dirtyBase;
    /// Base matrix for explosion ie. at center.
    float baseMatrix[16];
    /// Explosion to be triggered on destruction of the building.
    cExplosion explosionObject;
    /// Dimensions of the building.
    float buildingRooms[3];
public:
    cBuilding(int x, int y, int z, int rooms_x, int rooms_y, int rooms_z);
    
    virtual void damageByParticle(float* localpos, float damage, cObject* enactor = NULL);
    virtual float constrainParticle(float* worldpos, float radius = 0.0f, float* localpos = NULL, cObject* enactor = NULL);
    
    virtual void onSpawn();
    
    virtual void animate(float spf);
    virtual void transform();
    virtual void drawSolid();
    virtual void drawEffect();
};

struct cParticle;

/**
 * Models a set/cluster/field of random distributed small
 * environmental decal objects such as stones, flowers,
 * grass, dirt and debris.
 */
class cScatterfield : public cObject {
protected:
    static int sInstances;
    static std::map<int,long> sTextures;
public:
    enum CollisionType { CT_NONE, CT_CYLINDER, CT_SPHERE, CT_BOX };
    enum BillboardType { BT_NONE, BT_AROUND, BT_AXIS };
protected:
    std::list<cParticle*> decalParticles;
public:
    cScatterfield(float x, float y, float z, float radius, float density);
    virtual void drawEffect();
};


/**
 * Models an urban street-tile of different types.
 */
class cTile : public cObject {
public:
    /// Possible models of tile.
    enum Kinds {
        KIND_CONCRETE,
        KIND_ROAD_EASTWEST,
        KIND_ROAD_NORTHSOUTH,
        KIND_ROAD_JUNKTION,

        MAX_KINDS
    };
protected:
    /// Counts number of instances.
    static int sInstances;
    /// Contains shared texture(s) of all instances.
    static std::map<int,long> sTextures;
    
    //static long sDisplaylists[MAX_KINDS*2];
protected:
    /// This tile has the named appeareance (as in Kinds-enum).
    int tileKind;
public:

    /// Initialises a tile of a certain kind at the given tile-grid position.
    cTile(int x, int y, int z, int kind);
    virtual void drawSolid();
};


class cPadmap : public cObject {
public:
    /// Possible models of tile.
    enum Kinds {
        KIND_CONCRETE,
        KIND_ROAD_EASTWEST,
        KIND_ROAD_NORTHSOUTH,
        KIND_ROAD_JUNKTION,

        MAX_KINDS
    };
protected:
    /// Counts number of instances.
    static int sInstances;
    /// Contains shared texture(s) of all instances.
    static std::map<int,long> sTextures;
    static std::vector<int> sSpiral[2];
protected:
    int dims[2];
    float heights[16*16];
public:
    cPadmap(float x, float z);
    virtual float constrainParticle(float* worldpos, float radius = 0.0f, float* localpos = NULL, cObject* enactor = NULL);
    virtual void drawSolid();
};


#endif
