#include "cPlanetmap.h"

#include "psi3d/macros.h"
#include "psi3d/noise.h"
#include "cTree.h"

#include <GL/gl.h>
#define GL_GLEXT_PROTOTYPES
#include <GL/glext.h>

#include <ostream>

using namespace std;


int cPlanetmap::sInstances = 0;
std::vector<long> cPlanetmap::sTextures;


cPlanetmap::cPlanetmap() {
    sInstances++;
    if (sInstances == 1) {
        // Solid Ground
        {
            const float diffusion = 0.2;
            const float shift = 0.0;
            const int size = 64;
            unsigned char texels[size * size * size * 3];
            int samples = 3;
            float amt = 1.0f / (float)(samples);
            for (int l = 0; l < samples; l++) {
                unsigned char* texel = texels;
                loopijk(size, size, size) {
                    float f = 256.0f / (float)size;
                    float i_ = (rand()%100) * 0.01;
                    float j_ = (rand()%100) * 0.01;
                    float k_ = (rand()%100) * 0.01;
                    unsigned char r = 0, g = 0, b = 0;
                    float a = ((voronoi3((i+i_)*f,(j+j_)*f,(k+k_)*f, diffusion, shift) + 1.0f) * 0.5);
                    a = 0.3 + a * 0.8;
                    a = a > 1.0 ? 1.0 : a;
                    a = a < 0.0 ? 0.0 : a;
                    r = a * 255;

                    //float a = ((simplex3(i*f,j*f,k*f, 43) + 1.0f) * 0.5);
                    //r = (0.5 + 0.5 * (a - int(a))) * 255;
                    g = b = r ;
                    if (l == 0) {
                        *texel++ = r * amt;
                        *texel++ = g * amt;
                        *texel++ = b * amt;
                    } else {
                        *texel++ += r * amt;
                        *texel++ += g * amt;
                        *texel++ += b * amt;
                    }
                }
            }
            unsigned int texname;
            glBindTexture3D(texname, true, true, true, true, true, size, size, size, texels);
            sTextures.push_back(texname);
        }
        // Scattering 1
        {
            unsigned int texid;
            int w, h, bpp;
            unsigned char* texels = loadTGA("data/planet/background/desertplant.tga", &w, &h, &bpp);
            glBindTexture2D(texid, true, true, false, false, w, h, bpp, texels);
            delete texels;
            sTextures.push_back(texid);
        }
        // Scattering 2
        {
            unsigned int texid;
            int w, h, bpp;
            unsigned char* texels = loadTGA("data/planet/background/desertplant.tga", &w, &h, &bpp);
            glBindTexture2D(texid, true, true, false, false, w, h, bpp, texels);
            delete texels;
            sTextures.push_back(texid);
        }
        // Scattering 3
        {
            unsigned int texid;
            int w, h, bpp;
            unsigned char* texels = loadTGA("data/planet/background/desertplant.tga", &w, &h, &bpp);
            glBindTexture2D(texid, true, true, false, false, w, h, bpp, texels);
            delete texels;
            sTextures.push_back(texid);
        }
        sTextures.push_back(sTextures.back());
    }
    this->traceable->radius = 10;
    this->traceable->pos[0] = float_NAN;
    this->traceable->pos[1] = float_NAN;
    this->traceable->pos[2] = float_NAN;
    collideable = new rCollideable;
    addRole(COLLIDEABLE);
    damageable = new rDamageable;
    addRole(DAMAGEABLE);
    nameable->name = "PLANETMAP";
}



float cPlanetmap::constrainParticle(float* worldpos, float radius, float* localpos, cObject* enactor) {
    //if (enactor == NULL) return 0.0;
    //if (!enactor->hasRole(HUMANPLAYER)) return 0.0;
    float localpos_[3];
    vector_cpy(localpos_, worldpos);

    //assert(finitef(localpos_[0]));
    //assert(finitef(localpos_[1]));
    //assert(finitef(localpos_[2]));

    const int iterations = 24;
    const int samples = 16;
    float relax = 0.1;
    float maxdelta = 0;
    float color3f[3];
    int runs = 0;
    loopj(iterations) {
        runs++;
        //relax = 0.05 + 0.1 * (iterations - j) / (float)(iterations);
        relax = 0.01;
        //const float a = 2.0 * M_PI / (float) samples;
        //const float b = 0.5 * M_PI / (float) samples;
        float nearest2 = radius*radius;
        float near[] = {0, 0, 0};

        loopi(samples) {
            float x_ = localpos_[0];
            float y_ = localpos_[1];
            float z_ = localpos_[2];
            float alpha = 0;
            float beta = 0;
            if (i != 0) {
                //alpha = i * a;
                //beta = i * b;
                alpha = (rand()%62831)*0.0001f; // 2PI 0-360
                beta =  (rand()%15707)*0.0001f; // 0.5PI 0-90
            }
            x_ += sin(alpha) * sin(beta) * radius;
            y_ += -cos(beta) * radius;
            z_ += cos(alpha) * sin(beta) * radius;

            float h00 = getCachedHeight(x_, z_, color3f);
            float delta = y_ - h00;
            if (delta > 0) continue;
            if (h00 > localpos_[1]) {
                h00 = localpos_[1];
            }
            float v[] = {localpos_[0] - x_, localpos_[1] - h00, localpos_[2] - z_};
            float dist2 = vector_dot(v, v);
            if (dist2 < nearest2) {
                nearest2 = dist2;
                vector_cpy(near, v);
            }
        }
        float nearest = sqrt(nearest2);
        float delta = radius - nearest;
        if (delta > 0.01) {
            //vector_print(near);
            //cout << radius << " " << nearest << " " << delta << endl;

            vector_scale(near, near, 1.0f / (nearest+0.000001f));
            vector_scale(near, near, delta * relax);
            vector_add(localpos_, localpos_, near);
        } else if (j > 2) {
            break;
        }
        maxdelta = xmax(maxdelta, delta);
    }
    //cout << "it " << runs << endl;

    if (maxdelta > 0) {
        vector_cpy(worldpos, localpos_);
        if (localpos != NULL) vector_cpy(localpos, localpos_);
    }
    return maxdelta;
}

//float cPlanetmap::constrainParticle(float* worldpos, float radius, float* localpos, cObject* enactor) {
//    //if (enactor == NULL) return 0.0;
//    if (!enactor->hasRole(HUMANPLAYER)) return 0.0;
//    float localpos_[3];
//    vector_cpy(localpos_, worldpos);
//
//    //assert(finitef(localpos_[0]));
//    //assert(finitef(localpos_[1]));
//    //assert(finitef(localpos_[2]));
//
//    const int iterations = 24;
//    const int samples = 16;
//    float relax = 0.1;
//    float maxdelta = 0;
//    float color3f[3];
//    int runs = 0;
//    loopj(iterations) {
//        runs++;
//        //relax = 0.05 + 0.1 * (iterations - j) / (float)(iterations);
//        relax = 0.01;
//        //const float a = 2.0 * M_PI / (float) samples;
//        //const float b = 0.5 * M_PI / (float) samples;
//        float nearest2 = radius*radius;
//        float near[] = {0, 0, 0};
//
//        float furthest2 = -1;
//        float far[] = {0, 0, 0};
//
//        loopi(samples) {
//            float x_ = localpos_[0];
//            float y_ = localpos_[1];
//            float z_ = localpos_[2];
//            float alpha = 0;
//            float beta = 0;
//            if (i != 0) {
//                //alpha = i * a;
//                //beta = i * b;
//                alpha = (rand()%62831)*0.0001f; // 2PI 0-360
//                beta =  (rand()%15707)*0.0001f; // 0.5PI 0-90
//            }
//            x_ += sin(alpha) * sin(beta) * radius;
//            y_ += -cos(beta) * radius;
//            z_ += cos(alpha) * sin(beta) * radius;
////            float h00 = getCachedHeight(x_, z_, color3f);
////            float delta = y_ - h00;
////            if (delta > 0) continue;
////            if (h00 > localpos_[1]) {
////                h00 = localpos_[1];
////            }
////            float v[] = {localpos_[0] - x_, localpos_[1] - h00, localpos_[2] - z_};
////            float dist2 = vector_dot(v, v);
////            if (dist2 < nearest2) {
////                nearest2 = dist2;
////                vector_cpy(near, v);
////            }
//            float h00 = getCachedHeight(x_, z_, color3f);
//            float delta = y_ - h00;
//            if (delta > 0) continue;
//            float v[] = {localpos_[0] - x_, localpos_[1] - h00, localpos_[2] - z_};
//            float dist2 = vector_dot(v, v);
//            //cout << v[1] << " " << sqrt(dist2) << " locy " << localpos_[1] << " h00 " << h00 << "\n";
//            if (h00 > localpos_[1]) {
//                //h00 = localpos_[1];
//                if (dist2 >= furthest2) {
//                    furthest2 = dist2;
//                    vector_cpy(far, v);
//                }
//            } else {
//                if (dist2 < nearest2) {
//                    nearest2 = dist2;
//                    vector_cpy(near, v);
//                }
//            }
//        }
////        float nearest = sqrt(nearest2);
////        float delta = radius - nearest;
////        if (delta > 0.01) {
////            //vector_print(near);
////            //cout << radius << " " << nearest << " " << delta << endl;
////
////            vector_scale(near, near, 1.0f / (nearest+0.000001f));
////            vector_scale(near, near, delta * relax);
////            vector_add(localpos_, localpos_, near);
////        } else if (j > 2) {
////            break;
////        }
////        maxdelta = xmax(maxdelta, delta);
//        float delta = 0;
//        if (furthest2 >= 0) {
//            float furthest = sqrt(furthest2);
//            float delta = (radius + furthest);
//            cout << "furthest " << furthest << " delta " << delta << "\n";
//            {
//                vector_print(far);
//
//                // normalise
//                //vector_scale(far, far, 1.0f / (furthest+0.000001f));
//                //vector_scale(far, far, -delta * relax);
//                //vector_add(localpos_, localpos_, far);
//                //localpos_[1] -= 10;
//            }
//        } else {
////            float nearest = sqrt(nearest2);
////            float delta = radius - nearest;
////            if (delta > 0.01) {
////                //vector_print(near);
////                //cout << radius << " " << nearest << " " << delta << endl;
////
////                vector_scale(near, near, 1.0f / (nearest+0.000001f));
////                vector_scale(near, near, delta * relax);
////                vector_add(localpos_, localpos_, near);
////            } else if (j > 2) {
////                break;
////            }
//        }
//        maxdelta = xmax(maxdelta, delta);
//    }
//    //cout << "it " << runs << endl;
//
//    if (maxdelta > 0) {
//        vector_cpy(worldpos, localpos_);
//        if (localpos != NULL) vector_cpy(localpos, localpos_);
//    }
//    return maxdelta;
//}


/**
 * ~[0,1] normalized utah rock landscape.
 */
inline float getUtah(float x, float z, float* color3f) {
    float colors[][3] = {
        {0.40f, 0.30f, 0.20f},
        {0.70f, 0.55f, 0.25f},
        {0.80f, 0.70f, 0.40f},
        {0.40f, 0.30f, 0.20f},
        {0.70f, 0.55f, 0.25f},
        {0.80f, 0.70f, 0.40f},
        {0.70f, 0.55f, 0.25f},
        {0.80f, 0.70f, 0.40f},
    };

    float x_ = x;
    float z_ = z;
    const int seed2 = 17;
    const int seed3 = 23;
    // Prime periods => large smallest common multiple.
    const float p2 = 1.0f / 107.0f;
    const float p3 = 1.0f / 47.0f;
    float o2 = simplex3(50 + x_*p2, 0, 50 + z_*p2, seed2);
    float o3 = simplex3(10 + x_*p3, 0, 90 + z_*p3, seed3);

    float h = exp((o2 + 0.5f * o3)-0.6f) * 0.249f;

    h = (h > 0.6f) ? 0.6f : h;
    h = (h < 0.2f) ? 0.0f : h;

    int i = ((int) (h * 16.0f)) & 7;
    color3f[0] = colors[i][0];
    color3f[1] = colors[i][1];
    color3f[2] = colors[i][2];

    return h;
}

/**
 * ~[0,1] normalized desert sand dune landscape.
 */
inline float getDuneland(float x, float z, float* color3f) {
    const int seed1 = 117;
    // Prime periods => large smallest common multiple.
    const float p1 = 1.0f / 101.0f;

    float x_ = 90 + x * p1;// + 5*sin(z * PI_OVER_180*30);
    float z_ = 10 + 2.0f * z * p1;

    float o1 = simplex3(x_, 0, z_, seed1);

    const float a = 0.5f * p1;
    const float b = a * 0.5f;

    float o1_ = simplex3(x_ + a, 0, z_ + b, seed1);

    // Limit to about [0,1] range.
    float h = (xmin(o1, -o1_)*0.98f+0.96f);

    float c1[] = { 1.00f, 0.96f, 0.79f };
    float c0[] = { 0.86f, 0.78f, 0.59f };
    float alpha = h;
    float alpha_ = 1.0f - h;
    color3f[0] = c0[0] * alpha_ + c1[0] * alpha;
    color3f[1] = c0[1] * alpha_ + c1[1] * alpha;
    color3f[2] = c0[2] * alpha_ + c1[2] * alpha;

    return h;
}


/**
 * ~[0,1] normalized lavatic landscape.
 */
inline float getLavaland(float x, float z, float* color3f) {

    float power = 5;
    float wlen = 1.0f/ (2*power);
    float dx = power*simplex3(37 + x*wlen, 0, 97 + 2.0f*z*wlen, 241);
    float dz = power*simplex3(67 + x*wlen, 0, 7 + 2.0f*z*wlen, 21);
    x += dx;
    z += dz;

    const int seed1 = 117;
    // Prime periods => large smallest common multiple.
    const float p1 = 1.0f / 101.0f;

    float x_ = 90 + x * p1;
    float z_ = 10 + z * p1;

    float o1 = simplex3(x_, 0, z_, seed1);

    const float a = 0.5f * p1;
    const float b = a * 0.5f;

    float o1_ = simplex3(x_ + a, 0, z_ + b, seed1);

    // Limit to about [0,1] range.
    float h = 1.0f-(xmin(o1, -o1_)*0.98f+0.96f);

    //float diffusion = 0.9;
    //float shift = 0.5;
    //h = 1.0f-voronoi3(x_, 0, z_,diffusion,shift);

    float ash[] = { 0.3f, 0.3f, 0.35f };
    float lava[] = { 1.0f, 0.0f, 0.0f };
    float glow[] = { 1.1f, 1.1f, 1.1f };
    float ash_alpha = h*h;
    float lava_alpha = 0.25f*(2.0f-2.0f*h)*(2.0f-2.0f*h);
    float glow_alpha = (h < 0.2f) ? 1.2f-5.0f*h : 0.0f;
    color3f[0] = lava[0] * lava_alpha + ash[0] * ash_alpha + glow[0] * glow_alpha;
    color3f[1] = lava[1] * lava_alpha + ash[1] * ash_alpha + glow[1] * glow_alpha;
    color3f[2] = lava[2] * lava_alpha + ash[2] * ash_alpha + glow[2] * glow_alpha;

    return h;
}


inline float getSpikes(float x, float z, float* color3f) {
    int seed = 11;
    const int n = 4;
    float o = 0;
    const float p[] = { 1.0f/47.0f, 1.0f/23.0f, 1.0f/11.0f, 1.0f/ 7.0f };
    const float sx[] = { 231,123,45,31,13,7,3 };
    float factor = 0.9f;
    x *= 1.9f;
    z *= 1.9f;
    loopi(n) {
        o += factor * simplex3(sx[i] + x*p[i], 0, i*(i+1) + z*p[i], seed);
        factor *= 0.8f;
        seed = (seed+1)<<1;
    }

    float h = exp(o);

    h-= 0.6f+4;
    h = (h < 0.2f) ? 0.0f : h;
    h *= 1.0f;

    color3f[0] = color3f[1] = color3f[2] = h;

    return h;
}

inline float getStones(float x, float z, float* color3f) {
    int seed = 11;
    const int n = 4;
    float o = 0;
    const float p[] = { 1.0f/47.0f, 1.0f/23.0f, 1.0f/11.0f, 1.0f/ 7.0f };
    const float sx[] = { 231,123,45,31,13,7,3 };
    float factor = 0.9f;
    x *= 0.5f;
    z *= 0.5f;
    loopi(n) {
        o += factor * simplex3(sx[i] + x*p[i], 0, i*(i+1) + z*p[i], seed);
        factor *= 0.8f;
        seed = (seed+1)<<1;
    }

    float h = o;
    float t = 1.25f;
    h = t*o * sin(t*o)-0.9f;

    h-= 0.2f;
    h = (h < 0.2f) ? 0.0f : h;
    h *= 1.0f;

    color3f[0] = color3f[1] = color3f[2] = h;

    return h;
}


inline float getLavamashup(float x, float z, float* color3f) {
    const int seed1 = 11;
    const float p1 = 1.0f / 301.0f;
    float o1 = 0.5f + 0.5f * simplex3(37 + x*p1, 0, 97 + z*p1, seed1);

    float dunecolor[3];
    float utahcolor[3];
    //float dune = 12 * o1 * getDune(x, z, dunecolor);
    //float utah = 90 * getUtah(x,z,utahcolor);
    float dune = 12 * o1 * getLavaland(x, z, dunecolor);
    float utah = 5 * getSpikes(x, z, utahcolor);
    bool isdune = dune > utah;

    float h = (isdune ? dune : utah);

    float stonecolor[3];
    float stone = 5 * getStones(x, z, stonecolor);
    bool isstone = (stone > 0.001) && isdune;
    if (isstone) {
        h += stone;
    }

    if (isstone) {
        color3f[0] = stonecolor[0];
        color3f[1] = stonecolor[1];
        color3f[2] = stonecolor[2];
    } else if (isdune) {
        color3f[0] = dunecolor[0];
        color3f[1] = dunecolor[1];
        color3f[2] = dunecolor[2];
    } else {
        color3f[0] = utahcolor[0];
        color3f[1] = utahcolor[1];
        color3f[2] = utahcolor[2];
    }
    
    return h;
}


inline float getGrassland(float x, float z, float* color3f) {
    const int seed1 = 11;
    const float p1 = 1.0f / 101.0f;
    float o1 = simplex3(37 + x*p1, 0, 97 + z*p1, seed1);

    const int seed2 = seed1+43;
    const float p2 = 1.0f / 47.0f;
    float o2 = simplex3(71 + x*p2, 0, 19 + z*p2, seed2);

    float h = (0.9f * o1 + 0.5f * o2) * 0.5 + 0.5;

    const int seed3 = seed1+43;
    const float p3 = 1.0f / 17.0f;
    float o3 = simplex3(21 + x*p3, 0, 79 + z*p3, seed3);

    float top[] = { 0.59f, 0.80f, 0.58f };
    float bottom[] = { 0.85f, 0.93f, 0.58f };
    float splat[] = { 0.93f, 0.95f, 0.61f };
    float top_alpha = h;
    float bottom_alpha = 1.0f - h;
    float splat_alpha = (exp(o2 * 0.5 + 0.5))*0.1 + (exp(o3 * 0.5 + 0.5))*0.05;

    color3f[0] = top[0] * top_alpha + bottom[0] * bottom_alpha + splat[0] * splat_alpha;
    color3f[1] = top[1] * top_alpha + bottom[1] * bottom_alpha + splat[1] * splat_alpha;
    color3f[2] = top[2] * top_alpha + bottom[2] * bottom_alpha + splat[2] * splat_alpha;

    return h;
}


inline float getSnowland(float x, float z, float* color3f) {
    const int seed1 = 11;
    const float p1 = 1.0f / 101.0f;
    float o1 = simplex3(37 + x*p1, 0, 97 + z*p1, seed1);

    const int seed2 = seed1+43;
    const float p2 = 1.0f / 47.0f;
    float o2 = simplex3(71 + x*p2, 0, 19 + z*p2, seed2);

    float h = (0.9f * o1 + 0.5f * o2) * 0.5 + 0.5;

    const int seed3 = seed1+43;
    const float p3 = 1.0f / 17.0f;
    float o3 = simplex3(21 + (x+sin(z))*p3, 0, 79 + (z+sin(x))*p3, seed3);

    float top[] = { 0.86f, 0.89f, 0.89f };
    float bottom[] = { 0.98f, 0.98f, 0.98f };
    float splat[] = { 0.58f, 0.85f, 0.95f };
    float top_alpha = h;
    float bottom_alpha = 1.0f - h;
    float splat_alpha = (exp(o2 * 0.5 + 0.5))*0.1 + (exp(o3 * 0.5 + 0.5))*0.3;

    color3f[0] = top[0] * top_alpha + bottom[0] * bottom_alpha + splat[0] * splat_alpha;
    color3f[1] = top[1] * top_alpha + bottom[1] * bottom_alpha + splat[1] * splat_alpha;
    color3f[2] = top[2] * top_alpha + bottom[2] * bottom_alpha + splat[2] * splat_alpha;

    return h;
}

inline float getDirtland(float x, float z, float* color3f) {
    const int seed1 = 11;
    const float p1 = 1.0f / 101.0f;
    float o1 = simplex3(37 + x*p1, 0, 97 + z*p1, seed1);

    const int seed2 = seed1+43;
    const float p2 = 1.0f / 47.0f;
    float o2 = simplex3(71 + x*p2, 0, 19 + z*p2, seed2);

    float h = (0.9f * o1 + 0.5f * o2) * 0.5 + 0.5;

    const int seed3 = seed1+43;
    const float p3 = 1.0f / 17.0f;
    float o3 = simplex3(21 + x*p3, 0, 79 + z*p3, seed3);

    float top[] = { 0.59f, 0.50f, 0.18f };
    float bottom[] = { 0.75f, 0.73f, 0.28f };
    float splat[] = { 0.43f, 0.35f, 0.21f };
    float top_alpha = h;
    float bottom_alpha = 1.0f - h;
    float splat_alpha = (exp(o2 * 0.5 + 0.5))*0.1 + (exp(o3 * 0.5 + 0.5))*0.05;

    color3f[0] = top[0] * top_alpha + bottom[0] * bottom_alpha + splat[0] * splat_alpha;
    color3f[1] = top[1] * top_alpha + bottom[1] * bottom_alpha + splat[1] * splat_alpha;
    color3f[2] = top[2] * top_alpha + bottom[2] * bottom_alpha + splat[2] * splat_alpha;

    return h;
}


inline float continentalRidge(float x) {
    float a = pow(x*2.05f, 12);//1.1f * x;
    //float b = 0.5f * x + 0.26f;
    float b = 0.61f * x + 0.197f;
    float c = 0.51f;
    float d = pow(1.0f-fabs(x-0.57), 11);
    float h = xmax(d, xmin(a, xmin(b, c)));
    return h;
}

inline float getContinental(float x, float z, float* color3f) {
    int seed = 11;
    const int n = 8;
    float o = 0;
    const float p[] = { 1.0f/5003.0f, 1.0f/1001.0f, 1.0f/301.0f, 1.0f/47.0f, 1.0f/23.0f, 1.0f/11.0f, 1.0f/ 7.0f, 1.0f/3.0f };
    const float sx[] = { 231,123,45,31,13,7,3,1 };
    float factor = 0.9;
    x *= 0.1;
    z *= 0.1;
    loopi(n) {
        o += factor * simplex3(sx[i] + x*p[i], 0, i*(i+1) + z*p[i], seed);
        factor *= 0.8;
        seed = (seed+1)<<1;
    }

    float h = continentalRidge(o*0.5f+0.5f);

    float sea[] = { 0.1,0.1,0.3 };
    float beach[] = { 0.7, 0.7, 0.5 };
    //float grass[] = { 0.3, 0.7, 0.3 };
    //float peak[] = { 1,1,1 };
    float peak[] = { 0.4f,0.3f,0 };

    float sea_alpha = (1.0f - h)*(1.0f - h);
    float beach_alpha = (1.0f - xabs(h - 0.5))*(1.0f - xabs(h - 0.5));
    float peak_alpha = h*h;

    loopi(3) {
        color3f[i] = sea[i] * sea_alpha + beach[i] * beach_alpha + peak[i] * peak_alpha;
    };

    return h;
}



float cPlanetmap::getHeight(float x, float z, float* color3f) {
    float h = 0;

    if (1) {
        float grass[3] = {0,0,0};
        float snow[3] = {0,0,0};

        // Blending between types.
        const int seed1 = 121;
        const float p1 = 1.0f / 301.0f;
        float o1 = simplex3(7 + x*p1, 0, 19 + z*p1, seed1);
        o1 = sig(o1*28.0f);

        float h1 = 0;
        float h0 = 0;
        //if (o1 > 0.005) h1 = 1*getSnowland(x, z, snow);
        //if (o1 < 0.995) h0 = 10*getSnowland(x, z, grass);
        //if (o1 > 0.005) h1 = 1*getLavamashup(x, z, snow);
        //if (o1 < 0.995) h0 = 10*getLavaland(x, z, grass);
        if (o1 > 0.005) h1 = 1*getDuneland(x, z, snow);
        if (o1 < 0.995) h0 = 10*getDuneland(x, z, grass);
        h = (1.0f - o1) * h0 + o1 * h1 - 9;

        float color_[3];
        float h_ = 7*getStones(x,z,color_);
        if (h_ > 0) {
            color3f[0] = 1.0f-color_[0];
            color3f[1] = 1.0f-color_[1];
            color3f[2] = 1.0f-color_[2];
            h += h_;
        } else {
            color3f[0] = (1.0f-o1)*grass[0] + o1*snow[0];
            color3f[1] = (1.0f-o1)*grass[1] + o1*snow[1];
            color3f[2] = (1.0f-o1)*grass[2] + o1*snow[2];
        }

        // Large Scale Hills.
        const int seed2 = 97;
        const float p2 = 1.0f / 1001.0f;
        float o2 = simplex3(73 + x*p2, 0, 39 + z*p2, seed2);
        h += sig(o2*4-1.4) * 30;
    }

    if (0) {
        float f = 0.5;
        float cont[3];
        float hx = 15/f * (getContinental(x*f, z*f, cont) - 0.5f);
        //if (hx > 0.0) h += 2.0f*hx;
        //else h = hx;
        //h = hx+h;
        h = hx;
        color3f[0] = cont[0];
        color3f[1] = cont[1];
        color3f[2] = cont[2];
    }

    // Modification (Big City)
    {
        const float range = 70+65;
        const float height = 0.009f;
        const float inv_range = 1.0f / range;
        float x_ = x - 50;
        float z_ = z - 50;
        if (-range < x_ && x_ < +range && -range < z_ && z_ < +range) {
            float dx = xabs(x_);
            float dz = xabs(z_);
            float dy = 1.0 - xmax(dx,dz) * inv_range;
            dy = (dy < 0.25f) ? (dy * 4.0f) : 1.0f;
            h *= (1.0f - dy);
            h += height * dy;
        }
    }
    // Modification (Far City)
    {
        const float range = 60;
        const float height = 0.011f;
        const float inv_range = 1.0f / range;
        float x_ = x - 360;
        float z_ = z - 125;
        if (-range < x_ && x_ < +range && -range < z_ && z_ < +range) {
            float dx = xabs(x_);
            float dz = xabs(z_);
            float dy = 1.0 - xmax(dx,dz) * inv_range;
            dy = (dy < 0.5f) ? (dy * 2.0f) : 1.0f;
            h *= (1.0f - dy);
            h += height * dy;
        }
    }
    // Modification Space-Port (-260 -180)
    {
        const float range = 90;
        const float height = 0.011f;
        const float inv_range = 1.0f / range;
        float x_ = x + 260;
        float z_ = z + 180;
        if (-range < x_ && x_ < +range && -range < z_ && z_ < +range) {
            float dx = xabs(x_);
            float dz = xabs(z_);
            float dy = 1.0 - xmax(dx,dz) * inv_range;
            dy = (dy < 0.25f) ? (dy * 4.0f) : 1.0f;
            h *= (1.0f - dy);
            h += height * dy;
        }
    }

    // Modification Pyramid (-210 285)
    {
        const float range = 100;
        const float height = 0.011f;
        const float inv_range = 1.0f / range;
        //float x_ = x + 210;
        //float z_ = z - 285;
        float x_ = x + 140;
        float z_ = z - 210;
        if (-range < x_ && x_ < +range && -range < z_ && z_ < +range) {
            float dx = xabs(x_);
            float dz = xabs(z_);
            float dy = 1.0 - xmax(dx,dz) * inv_range;
            dy = (dy < 0.25f) ? (dy * 4.0f) : 1.0f;
            h *= (1.0f - dy);
            h += height * dy;
        }
    }

    return h;
} // getHeight



float cPlanetmap::getCachedHeight(float x, float z, float* color3f) {
    // nan?
    if (x != x || z != z) return 0;
#if 0
    return getHeight(x, z, color3f);
#else
    const long patchsize = 8; // 2^patchsize
    const long tilesize = 10; // 2^tilesize
    const long tilesize_ = (1L << tilesize)-1;
    const long detailfactor = (1L << (tilesize - patchsize));
    long tx = ((long) x) >> patchsize;
    long tz = ((long) z) >> patchsize;
    tx &= 0xFFFF;
    tz &= 0xFFFF;
    unsigned long key = (tx << 16) | tz;

    sPatch* patch = patches[key]; // This may already add a new key.
    if (patch == NULL) {
        bool recycled = false;
        if (patches.size() > 24) {
            recycled = true;
            unsigned long mintouches = -1;
            sPatch* minpatch = NULL;
            std::map<unsigned long,sPatch*>::iterator mini;
            std::map<unsigned long,sPatch*>::iterator i;
            for (i = patches.begin(); i != patches.end(); i++) {
                sPatch* patch_ = i->second;
                if (patch_ == NULL) continue;
                if (patch_->touches < mintouches) {
                    mintouches = patch_->touches;
                    minpatch = patch_;
                    mini = i;
                }
            }
            patches.erase(mini);
            patch = minpatch;
        } else {
            patch = new sPatch();
            assert(patch != NULL);
        }
        patch->touches = 0;
        patch->key = key;
        patches[patch->key] = patch;
        float* ptr = patch->heightcolor;

        const unsigned long totaltilesize = (1UL << (tilesize + tilesize)) << 2;
        for (unsigned long i = 0; i < totaltilesize; i++) {
            *ptr++ = float_NAN;
        }
        
        /*
        tx = ((long) x) >> patchsize;
        tz = ((long) z) >> patchsize;
        loopj(tilesize_+1) {
            loopi(tilesize_+1) {
                loopk(4) {
                *ptr++ = float_NAN;
                //*ptr++ = getHeight((tx << patchsize)+i*0.5f, (tz << patchsize)+j*0.5f, 4);
                }
            }
        }
        */
        ptr = patch->heightcolor;
        cout << "Patch=" << (recycled?"recycled":"new") << ", total=" << patches.size() << ", key=" << key << ", ptr=" << ptr << "\n";
    }
    patch->touches++;

    unsigned long fx = ((unsigned long)(fabs(x) * detailfactor));
    unsigned long fz = ((unsigned long)(fabs(z) * detailfactor));
    fx  &= tilesize_;
    fz  &= tilesize_;

    float* h = &(patch->heightcolor[(((fx << tilesize) | fz))<<2]);

    if (*h != *h) {
        float h_ = *h++ = getHeight(x, z, color3f);
        *h++ = color3f[0];
        *h++ = color3f[1];
        *h = color3f[2];
        return h_;
    } else {
        float h_ = *h++;
        color3f[0] = *h++;
        color3f[1] = *h++;
        color3f[2] = *h;
        return h_;
    }
#endif
} // getCachedHeight


int T = 0;
int detail = 0;

void cPlanetmap::drawSolid() {

    // Get current Camera-Matrix.
    float m[16];
    glGetFloatv(GL_MODELVIEW_MATRIX, m);
    // Position from 4th Col.
    float p[3];
    loop3i(p[i] = m[12 + i]);
    // Reset Position on 4th Col.
    loop3i(m[12 + i] = 0);
    // Apply transpose-inverse to position.
    float n[16];
    memcpy(n, m, sizeof (float) *16);
    matrix_transpose(m);
    matrix_apply2(m, p);

    //cout << "here!"<< endl;
    if (T % 100 == 0) detail = (detail + 1) % 4;
    detail = 5;

    float no[3];
    float relheight = p[1] + this->getCachedHeight(-p[0],-p[2], no);

    int hscale = -relheight * 0.05f;
    hscale = (hscale < 2) ? hscale : 1;
    if (0) detail -= hscale;
    T++;
    glPushMatrix();
    {
        glPushAttrib(GL_ALL_ATTRIB_BITS);
        {
            glDisable(GL_TEXTURE_2D);
            glEnable(GL_LIGHTING);
            glEnable(GL_FOG);


            if (1 && !(p[0] != p[0] && p[2] != p[2])) {
                // not nan position
                glDisable(GL_TEXTURE_2D);
                glEnable(GL_TEXTURE_3D);
                glEnable(GL_FOG);
                glBindTexture(GL_TEXTURE_3D, sTextures[0]);
                float step = 64;
                int s = 6*step * 16;

                loopi(detail) {
                    step *= 0.5;
                    s >>= 1;
                }
                detail = 2;
                float x = -((int) (p[0] / step)) * step;
                float z = -((int) (p[2] / step)) * step;
                assert(!(x!=x));
                assert(!(z!=z));
                //float c[3];
                float ts = 0.1;
                //cout << x << " " << z << endl;
                float color3f[4];
                for (float i = x - s; i < x + s; i += step) {
                    glBegin(GL_TRIANGLE_STRIP);
                    glNormal3f(0,1,0);
                    for (float j = z - s; j < z + s; j += step) {

                        const float dext = 180*180;
                        float d2 = ((j + p[2])*(j + p[2]) + (i + step + p[0])*(i + step + p[0]));
                        float alpha = (d2 < dext) ? 1 : 1.0f/(1.0f + 0.002f*(d2-dext));
                        color3f[3] = alpha;

                        float h0 = getCachedHeight(i + step, j, color3f);
                        glColor4fv(color3f);
                        glTexCoord3f((i + step)*ts, h0*ts, j*ts);
                        glVertex3f(i + step, h0, j);

                        d2 = ((j + p[2])*(j + p[2]) + (i + p[0])*(i + p[0]));
                        alpha = (d2 < dext) ? 1 : 1.0f/(1.0f + 0.002f*(d2-dext));
                        color3f[3] = alpha;

                        float h1 = getCachedHeight(i + 0, j, color3f);
                        glColor4fv(color3f);
                        glTexCoord3f((i + 0)*ts, h1*ts, j*ts);
                        glVertex3f(i + 0, h1, j);
                    }
                    glEnd();
                }
            }
        }
        glPopAttrib();
    }
    glPopMatrix();

    // Reset Touches for LRU
    for (std::map<unsigned long,sPatch*>::iterator i = patches.begin(); i != patches.end(); i++) {
        //cout << "A2" << endl;
        sPatch* patch_ = i->second;
        if (patch_ == NULL) continue;
        patch_->touches = 0;
    }
}


void cPlanetmap::drawEffect() {
    // Pseudorandom Permutation.
    static unsigned char perms[3*256];
    static bool init = true;
    if (init) {
        init = false;
        unsigned char b = 131;
        loopi(3*256) {
            if ((i & 255) == 255) {
                perms[i] = 255;
            } else {
                perms[i] = b = LFSR8(b);
            }
        }
    }

    // Get current Camera-Matrix.
    float m[16];
    glGetFloatv(GL_MODELVIEW_MATRIX, m);
    // Position from 4th Col.
    float p[3];
    loop3i(p[i] = m[12 + i]);
    // Reset Position on 4th Col.
    loop3i(m[12 + i] = 0);
    // Apply transpose-inverse to position.
    matrix_transpose(m);
    matrix_apply2(m, p);

    // Construct Billboarding Matrix.
    float n[16];
    glGetTransposeInverseRotationMatrix(n);
    vector_set(&n[4], 0,1,0);
    vector_cross(&n[0], &n[4], &n[8]);
    vector_cross(&n[8], &n[0], &n[4]);

    float* look = &n[8];
    int xlook = -((look[0]<0)?+1:-1);
    int zlook = -((look[2]<0)?+1:-1);

    // nan?
    if (p[0] != p[0] || p[2] != p[2]) return;

    glPushAttrib(GL_CURRENT_BIT | GL_ENABLE_BIT | GL_TEXTURE_BIT);
    {
        glDisable(GL_CULL_FACE);
        glEnable(GL_LIGHTING);
        glEnable(GL_TEXTURE_2D);
        glEnable(GL_FOG);
        glColor4f(0.8f, 0.8f, 0.8f, 1);
        glNormal3f(0,1,0);

        const float step = 16;
        const float near = (step*2)*(step*2);
        const float s = 8 * step;
        const float x = -((int) (p[0] / step)) * step;
        const float z = -((int) (p[2] / step)) * step;
        //cout << x << " " << z << endl;
        float  color3f[3];

        // Loop in positive or negative x direction (painters algorithm).
        float i = x-s*xlook;
        while (true) {
            if (fabs(x+s*xlook - i) < 0.0001f) break;

            float x_ = i;

            // Loop in positive or negative z direction (painters algorithm).
            float j = z-s*zlook;
            while (true) {
                if (fabs(z+s*zlook - j) < 0.0001f) break;

                float z_ = j;
                float dx = (-p[0]-x_-0.5f*step);
                float dz = (-p[2]-z_-0.5f*step);
                float dist2 = dx*dx + dz*dz;
                dist2 = (dist2 < near) ? near : dist2;
                float dist2_inv = 1.0f / (1.0f+0.0009f*dist2); // 0.0003 original
                float densityfactor = dist2_inv;

                unsigned char a = (char) x_;
                unsigned char b = (char) z_;
                unsigned char key = permutation2d(a,b);
                //int density = key * 1.5f * 0.015f * densityfactor; // original
                int density = (key+70) * 1.0f * 0.015f * densityfactor;
                //cout << density << endl;

                // Load LFSR using position, key and some scrambling.
                unsigned short lfsr16 = (((unsigned short)(a^key)) << 8) | (b^key);
                lfsr16 = (lfsr16 == 0) ? 1 : lfsr16;
                lfsr16 ^= 0x7493;

                loopi(density) {
                    glPushMatrix();
                    {
                        lfsr16 = LFSR16(lfsr16);
                        a = lfsr16;
                        b = lfsr16 >> 8;

                        float x__ = x_ + step * 0.003906f * a;
                        float z__ = z_ + step * 0.003906f * b;
                        float h = getCachedHeight(x__, z__, color3f);

                        lfsr16 = LFSR16(lfsr16);
                        unsigned char kind = lfsr16;
                        unsigned char rot = lfsr16 >> 8;

                        //lfsr16 = LFSR16(lfsr16);
                        //unsigned char size = lfsr16;
                        unsigned char size = rot;

                        const int under = 0;
                        if (i < under) {
                            int age = density - i;

                            glTranslatef(x__, h-0.2, z__);
                            glDisable(GL_TEXTURE_2D);
                            //GLuint dlist = cTree::compileTreeDisplaylist(1230+(b&2), c, 2+(a&1));
                            //GLuint dlist = cTree::compileTreeDisplaylist(1230+(b&2), 2, 2+(a&1));
                            GLuint dlist = cTree::compileTreeDisplaylist(1230+(b&2), 2, 2+age);
                            glCallList(dlist);
                            glEnable(GL_TEXTURE_2D);
                            glColor4f(1, 1, 1, 1);
                            glNormal3f(0,1,0);

                        } else {
                            unsigned char tex = (kind >> 4) & 3;

                            glTranslatef(x__, h-0.2, z__);
                            glMultMatrixf(n);
                            //glRotatef(rot*0.351563f, 0, 1, 0);
                            float s = 0.6 + 2.1*(size * 0.003906f);
                            glScalef(1*s,0.65*s,1*s);
                            //glAxis(0.9);
                            glXCenteredTextureSquare(sTextures[tex + 1]);
                            //glRotatef(90, 0, 1, 0);
                            //glXCenteredTextureSquare(sTextures[tex + 1]);
                        }
                    }
                    glPopMatrix();
                } // loopi density

                j += step*zlook;
            } // while j z

            i += step*xlook;
        } // while i x
    }
    glPopAttrib();
}


