/* 
 * File:     cPlanetmap.h
 * Project:  LinWarrior 3D
 * Home:     hackcraft.de
 *
 * Created on 18. Juli 2010, 21:39
 */

#ifndef PLANETMAP_H
#define	PLANETMAP_H

class cPlanetmap;

#include "cObject.h"

/**
 * Procedural Landscape Rendering and Collision.
 */
class cPlanetmap : public cObject {
public:
    /// Instance counter.
    static int sInstances;
    /// Instance shared Textures.
    static std::vector<long> sTextures;
public:
    /// LRU Surface Cache-Tile.
    struct sPatch {
        // Number of accesses last frame.
        unsigned long touches;
        // Cache key - spatial index.
        unsigned long key;
        // Cached Surface Data.
        float heightcolor[1024*1024*4];
    };
    /// LRU Surface Cache.
    std::map<unsigned long,sPatch*> patches;
public:
    cPlanetmap();
    /// Readjust particle position by making multiple downward-hemispherical checks.
    virtual float constrainParticle(float* worldpos, float radius, float* localpos, cObject* enactor);
    /// Calculate Height and Color of the xz position.
    inline float getHeight(float x, float z, float* color3f);
    /// Retrieves Height and Color of the xz position and calculates on demand.
    inline float getCachedHeight(float x, float z, float* color3f);
    /// Draw Landscape surrounding the current camera position.
    virtual void drawSolid();
    /// Draw Decals surrounding the current camera position.
    virtual void drawEffect();
};

#endif	/* PLANETMAP_H */

