#include "cWeapon.h"

#include "cObject.h"
#include "psi3d/macros.h"

#include <GL/gl.h>
#include <AL/al.h>
#include <AL/alut.h>

#include <sstream>
using std::stringstream;

#include <iostream>
using std::cout;
using std::endl;

#define DETAIL 0

#define WEAPONSOUND true
#define EXPLOSIONSOUND true

//--------------------------------

int cWeapon::sInstances = 0;
std::map<int,long> cWeapon::sTextures;

cWeapon::cWeapon() {
    sInstances++;
    if (sInstances == 1) {
        // Haze Texture
        if (1) {
            cout << "Generating Haze..." << endl;
            int w = 1 << 7;
            int h = w;
            int bpp = 4;
            unsigned char* texels = new unsigned char[((unsigned long)w)*h*bpp];
            unsigned char* p = texels;
            loopij(h,w) {

                float dx = (((float)j / (float)w) - 0.5f) * 2.0f;
                float dy = (((float)i / (float)h) - 0.5f) * 2.0f;
                float r = sqrt(dx*dx+dy*dy);
                float r_ = xmin(1, r);
                float a = 1.0f-pow(r_, 1.7);

                const float f = 2*2.0f;
                float v3f[] = { f * (float)j / (float)w, 0, f * (float)i / (float)h };
                float c = sig(simplex3(43+v3f[0],23+v3f[1],11+v3f[2])*4);
                float c4f[4] = { c, c, c, c };
                *p++ = 255*c4f[2];
                *p++ = 255*c4f[1];
                *p++ = 255*c4f[0];
                *p++ = 255*c4f[3]*a;
            }
            bool soft = true;
            bool repeat = true;
            unsigned int texname;
            glBindTexture2D(texname, true, soft, repeat, repeat, w, h, bpp, texels);
            sTextures[0] = texname;
            delete texels;
        }
    }

    weaponMountfv = weaponBasefv = NULL;
    weaponOwner = NULL;
    weaponScale = 1.0f;
    timeReloading = 0.0f;
    timeReadying = 0.0f;
    timeFiring = 0.0f;
    clipSize = 8;
    depotSize = 5;
    remainingAmmo = clipSize;
    remainingClips = depotSize;

    soundSource = -1;
}

void cWeapon::playSource() {
    if (alIsSource(soundSource)) {
        alSourcePlay(soundSource);
    }
}

void cWeapon::playSourceIfNotPlaying() {
    if (alIsSource(soundSource)) {
        ALint playing;
        alGetSourcei(soundSource, AL_SOURCE_STATE, &playing);
        if (playing != AL_PLAYING)
            alSourcePlay(soundSource);
    }
}

void cWeapon::transform() {
    { // Current Pose-Matrix and Sound-Source-Position update.
        glPushMatrix();
        {
            glLoadIdentity();
            if (weaponBasefv != NULL) glMultMatrixf(weaponBasefv);
            else if (weaponOwner) glTranslatef(weaponOwner->traceable->pos[0], weaponOwner->traceable->pos[1], weaponOwner->traceable->pos[2]);
            if (weaponMountfv) glMultMatrixf(weaponMountfv);
            glGetFloatv(GL_MODELVIEW_MATRIX, weaponPosef);
        }
        glPopMatrix();
        if (alIsSource(soundSource)) {
            alSourcefv(soundSource, AL_POSITION, &weaponPosef[12]);
        }
    }
}

int cWeapon::damageByParticle(float* worldpos, float radius, int roles, float damage) {
    roles = 0;//(1UL << cObject::DAMAGEABLE) | (1UL << cObject::COLLIDEABLE);
    float scaled_damage = damage * weaponScale;
    // Roles to test for.
    std::set<OID> test;
    int damaged = 0;
    float maxrange = 25;
    float worldpos_[3];
    vector_cpy(worldpos_, worldpos);
    std::list<cObject*>* range = cWorld::instance->filterByRange(weaponOwner, worldpos, 0.0f, maxrange, -1, NULL);
    //cout << "damageByParticle: "  << cWorld::instance->getNames(range) << endl;
    if (!range->empty()) {

        foreach(i, *range) {
            cObject* object = *i;
            if (! object->allRoles(&test) ) continue;
            float localpos_[3];
            float depth = object->constrainParticle(worldpos_, radius, localpos_, NULL);
            //cout << object->nameable->name << " depth: " << depth << endl;
            if (depth == 0) continue;
            if (object->hasRole(DAMAGEABLE)) {
                damaged++;
                object->damageByParticle(localpos_, scaled_damage, weaponOwner);
            }
            break;
        }
    }
    delete range;

    return damaged;
}

//--------------------------------

cMachineGun::cMachineGun() {
    //cout << "cMachineGun()\n";
    clipSize = 75;
    depotSize = 2;
    remainingAmmo = clipSize;
    remainingClips = depotSize;
    if (WEAPONSOUND) {
        //ALuint buffer = alutCreateBufferHelloWorld();
        ALuint buffer = alutCreateBufferFromFile("data/weapon/mgun/machinegun.wav");
        if (buffer == AL_NONE) throw "Could not load machinegun.wav";
        alGenSources(1, &soundSource);
        if (alGetError() != AL_FALSE || !alIsSource(soundSource)) throw "sorry, could not create audio source for machine gun";
        alSourcei(soundSource, AL_BUFFER, buffer);
        alSourcef(soundSource, AL_PITCH, 1.0f);
        alSourcef(soundSource, AL_GAIN, 1.5f);
        alSourcei(soundSource, AL_LOOPING, AL_FALSE);
    }
}

void cMachineGun::fire(OID target) {
    if (!ready()) return;

    if (remainingAmmo > 0) {
        remainingAmmo--;
        if (remainingAmmo == 0 && remainingClips != 0) {
            timeReloading = 2.5;
        } else {
            timeReloading = 0.5;
        }
    }

    float* source = weaponPosef;
    float* pos = &source[12];

    {
        cParticle* s = new cParticle();
        assert(s != NULL);
        s->fuel = 4;

        vector_cpy(s->pos, pos);

        float nrm[3];
        float pos2[] = {0, +3, 0};
        matrix_apply2(source, pos2);
        vector_sub(nrm, pos2, s->pos);

        vector_scale(s->vel, nrm, 1.4);
        const float variation = 0.01;
        loopi(3) s->vel[i] += (50 - (rand() % 100)) * variation;
        castoffParticles.push_back(s);
    }


    {
        cParticle* s = new cParticle();
        assert(s != NULL);
        s->fuel = 4;

        vector_cpy(s->pos, pos);

        float nrm[3];
        float pos2[] = {0, 0, -1};
        matrix_apply2(source, pos2);
        vector_sub(nrm, pos2, s->pos);

        vector_scale(s->vel, nrm, 200);
        const float variation = 0.03;
        loopi(3) s->vel[i] += (50 - (rand() % 100)) * variation;
        shrapnelParticles.push_back(s);
    }

    playSourceIfNotPlaying();
}

void cMachineGun::animate(float spf) {
    for (std::list<cParticle*>::iterator i = shrapnelParticles.begin(); i != shrapnelParticles.end();) {
        float damage = 10;
        cParticle* s = *i++;
        float dt = spf * 0.2f;
        int damaged = 0;

        loopi(5) {
            s->pos[0] += s->vel[0] * dt;
            s->pos[1] += s->vel[1] * dt;
            s->pos[2] += s->vel[2] * dt;
            damaged = this->damageByParticle(s->pos.data(), 0.2f, 0, damage);
            if (damaged != 0) break;
        }
        s->fuel -= spf;
        if (s->fuel < 0 || damaged != 0) {
            shrapnelParticles.remove(s);
            delete s;
        }
    }

    foreachNoInc(i, castoffParticles) {
        cParticle* s = *i++;
        s->vel[1] += cWorld::instance->mGravity[1] * spf;
        s->pos[0] += s->vel[0] * spf;
        s->pos[1] += s->vel[1] * spf;
        s->pos[2] += s->vel[2] * spf;
        s->fuel -= spf;
        if (s->fuel < 0) {
            castoffParticles.remove(s);
            delete s;
        }
    }

    foreachNoInc(i, damageParticles) {
        cParticle* s = *i++;
        s->pos[0] += s->vel[0] * spf;
        s->pos[1] += s->vel[1] * spf;
        s->pos[2] += s->vel[2] * spf;
        s->fuel -= spf;
        if (s->fuel < 0) {
            damageParticles.remove(s);
            delete s;
        }
    }

    timeReloading -= spf;
    if (timeReloading < 0) {
        timeReloading = 0.0f;
        if (remainingAmmo == 0 && remainingClips != 0) {
            remainingAmmo = clipSize;
            if (remainingClips > 0) remainingClips--;
        }
    }
}

void cMachineGun::drawSolid() {
    glPushAttrib(GL_ALL_ATTRIB_BITS);
    {
        glDisable(GL_CULL_FACE);
        glPushMatrix();
        {
            glMultMatrixf(weaponPosef);
            glRotatef(-90, 1, 0, 0);

            // Scale
            float scale = weaponScale;
            glScalef(scale, scale, scale);

            glColor4f(0.2, 0.3, 0.2, 1.0);
            glPushMatrix();
            {
                glScalef(0.1, 0.14, 0.12);
                glCenterUnitBlock();
            }
            glPopMatrix();

            if (this->ready() == 0 && remainingAmmo != 0) glRotatef(this->weaponOwner->base->seconds * 8 * 90, 0, 1, 0);

            glColor4f(0.3, 0.3, 0.3, 1.0);
            glPushMatrix();
            {
                glTranslatef(0, 0.75, 0);
                glScalef(0.065, 0.75, 0.065);
                glCenterUnitCylinder(7);

            }
            glPopMatrix();

            glColor4f(0.1, 0.1, 0.1, 1.0);
            int n = 3;

            loopi(n) {
                glPushMatrix();
                {
                    glRotatef(i * 360 / n, 0, 1, 0);
                    glTranslatef(0.075, 0.8, 0);
                    glScalef(0.03, 0.8, 0.03);
                    glCenterUnitCylinder(7);
                }
                glPopMatrix();
            }
        }
        glPopMatrix();
    }
    glPopAttrib();
}

void cMachineGun::drawEffect() {
    glPushAttrib(GL_ALL_ATTRIB_BITS);
    {

        glLineWidth(2);
        glDisable(GL_CULL_FACE);
        glDisable(GL_LIGHTING);
        glDisable(GL_TEXTURE_2D);
        glBlendFunc(GL_SRC_ALPHA, GL_ONE);
        glDepthMask(GL_FALSE);

        foreachNoInc(i, shrapnelParticles) {
            cParticle* s = *i++;
            const float len1 = 0.01;
            const float len2 = 0.02;
            glBegin(GL_LINE_STRIP);
            glColor4f(1, 1, 1, 1);
            glVertex3fv(s->pos.data());
            glColor4f(1, 1, 0, 1);
            glVertex3f(s->pos[0] - s->vel[0] * len1, s->pos[1] - s->vel[1] * len1, s->pos[2] - s->vel[2] * len1);
            glColor4f(1, 0, 0, 0.3);
            glVertex3f(s->pos[0] - s->vel[0] * len2, s->pos[1] - s->vel[1] * len2, s->pos[2] - s->vel[2] * len2);
            glEnd();
        }

        foreachNoInc(i, castoffParticles) {
            cParticle* s = *i++;
            const float len2 = 0.02;
            glBegin(GL_LINE_STRIP);
            glColor4f(1, 1, 1, 1);
            glVertex3fv(s->pos.data());
            //glColor4f(1, 1, 0, 1);
            //glVertex3f(s->pos[0] - s->vel[0] * len1, s->pos[1] - s->vel[1] * len1, s->pos[2] - s->vel[2] * len1);
            glColor4f(1, 0, 0, 0.3);
            glVertex3f(s->pos[0] - s->vel[0] * len2, s->pos[1] - s->vel[1] * len2, s->pos[2] - s->vel[2] * len2);
            glEnd();
        }

        float n[16];
        glGetTransposeInverseRotationMatrix(n);

        foreachNoInc(i, damageParticles) {
            cParticle* s = *i++;
            glPushMatrix();
            {
                glTranslatef(s->pos[0], s->pos[1], s->pos[2]);
                glScalef(0.6, 0.6, 0.8);
                glMultMatrixf(n);
                glColor4f(1, 0, 0, 0.7);
                glPrint(((std::string*)s->data)->c_str());
            }
            glPopMatrix();
        }

    }
    glPopAttrib();
}

void cMachineGun::drawHUD() {
    float a = 0.9;
    float b = 0.6;

    // Ammo
    float i = 0.1;
    float j = 0.9;
    float r = i + (remainingAmmo / (float) clipSize) * (j - i);
    glBegin(GL_QUADS);
    glColor4f(1, 1, 0, b);
    glVertex3f(i, j, 0.0);
    glVertex3f(i, i, 0.0);
    glColor4f(1, 0, 0, b);
    glVertex3f(r, i, 0.0);
    glVertex3f(r, j, 0.0);
    glEnd();

    // Clips
    float l = 0.1;
    float h = 0.15;
    r = i + (remainingClips / (float) depotSize) * (j - i);
    glBegin(GL_QUADS);
    glColor4f(0, 0, 0, a);
    glVertex3f(l, i + h, 0.0);
    glVertex3f(l, i, 0.0);
    glColor4f(1, 1, 1, a);
    glVertex3f(r, i, 0.0);
    glVertex3f(r, i + h, 0.0);
    glEnd();

    glBegin(GL_LINES);
    glColor4f(1, 1, 0, b);
    glVertex3f(0.25, 0.5, 0.0);
    glColor4f(1, 1, 1, a);
    glVertex3f(0.35, 0.5, 0.0);
    glColor4f(1, 1, 0, b);
    glVertex3f(0.45, 0.5, 0.0);
    glColor4f(1, 1, 1, a);
    glVertex3f(0.55, 0.5, 0.0);
    glColor4f(1, 1, 0, b);
    glVertex3f(0.65, 0.5, 0.0);
    glColor4f(1, 1, 1, a);
    glVertex3f(0.75, 0.5, 0.0);
    glEnd();
}


//--------------------------------

cPlasmaGun::cPlasmaGun() {
    clipSize = 19;
    depotSize = 9;
    remainingAmmo = clipSize;
    remainingClips = depotSize;

    if (WEAPONSOUND) {
        //ALuint buffer = alutCreateBufferHelloWorld();
        ALuint buffer = alutCreateBufferFromFile("data/weapon/pgun/plasmagun.wav");
        assert(buffer != AL_NONE);
        alGenSources(1, &soundSource);
        alSourcei(soundSource, AL_BUFFER, buffer);
        alSourcef(soundSource, AL_PITCH, 1.0f);
        alSourcef(soundSource, AL_GAIN, 1.0f);
        alSourcei(soundSource, AL_LOOPING, AL_FALSE);
    }
}

void cPlasmaGun::fire(OID target) {
    if (!ready()) return;

    if (remainingAmmo > 0) {
        remainingAmmo--;
        if (remainingAmmo == 0 && remainingClips != 0) {
            timeReloading = 2.5;
        } else {
            timeReloading = 0.3;
        }
    }

    float* source = weaponPosef;

    cParticle* s = new cParticle();
    assert(s != NULL);
    s->fuel = 3;

    float* pos = &source[12];
    vector_cpy(s->pos, pos)

            float nrm[3];
    float pos2[] = {0, 0, -1};
    matrix_apply2(source, pos2);
    vector_sub(nrm, pos2, s->pos);

    vector_scale(s->vel, nrm, 50);
    shrapnelParticles.push_back(s);

    playSourceIfNotPlaying();
}

void cPlasmaGun::animate(float spf) {

    foreachNoInc(i, shrapnelParticles) {
        cParticle* s = *i++;
        s->pos[0] += s->vel[0] * spf;
        s->pos[1] += s->vel[1] * spf;
        s->pos[2] += s->vel[2] * spf;
        s->fuel -= spf;
        if (s->fuel < 0) {
            shrapnelParticles.remove(s);
            delete s;
        } else {
            float radius = 0.2;
            int roles = 0;
            float damage = 13;
            int damaged = this->damageByParticle(s->pos.data(), radius, roles, damage);
            if (damaged) {
                shrapnelParticles.remove(s);
                delete s;
            }
        }
    }

    timeReloading -= spf;
    if (timeReloading < 0) {
        timeReloading = 0.0f;
        if (remainingAmmo == 0 && remainingClips != 0) {
            remainingAmmo = clipSize;
            if (remainingClips > 0) remainingClips--;
        }
    }
}

void cPlasmaGun::drawSolid() {
    glPushAttrib(GL_ALL_ATTRIB_BITS);
    {
        glDisable(GL_CULL_FACE);
        glPushMatrix();
        {
            glMultMatrixf(weaponPosef);
            glRotatef(-90, 1, 0, 0);

            // Scale
            float scale = weaponScale;
            glScalef(scale, scale, scale);

            if (this->ready() == 0 && remainingAmmo != 0) {
                glDisable(GL_LIGHTING);
                glColor4f(0.2, 0.2, 0.3, 1);
                glPushMatrix();
                {
                    glScalef(0.02, 0.02, 0.02);
                    glTranslatef(0, (0.5 + 0.5 * sin(this->weaponOwner->base->seconds * 2 * M_PI))* 1.7 / 0.02, 0);
                    glRotatef(this->weaponOwner->base->seconds * 2 * 360, 1, 3, 7);
                    glUnitBlock();
                }
                glPopMatrix();
                glColor4f(0.2, 0.3, 0.2, 1);
                glPushMatrix();
                {
                    glScalef(0.02, 0.02, 0.02);
                    glTranslatef(0, (0.5 + 0.5 * cos(this->weaponOwner->base->seconds * 2 * M_PI))* 1.7 / 0.02, 0);
                    glRotatef(this->weaponOwner->base->seconds * 2 * 360, 1, 3, 7);
                    glUnitBlock();
                }
                glPopMatrix();
                glColor4f(0.3, 0.2, 0.2, 1);
                glPushMatrix();
                {
                    glScalef(0.02, 0.02, 0.02);
                    glTranslatef(0, (0.5 - 0.5 * sin(this->weaponOwner->base->seconds * 2 * M_PI))* 1.7 / 0.02, 0);
                    glRotatef(this->weaponOwner->base->seconds * 2 * 360, 1, 3, 7);
                    glUnitBlock();
                }
                glPopMatrix();
            }

            glEnable(GL_LIGHTING);
            glColor4f(0.2, 0.2, 0.3, 1.0);
            glPushMatrix();
            {
                glScalef(0.1, 0.14, 0.12);
                glCenterUnitBlock();
            }
            glPopMatrix();

            glColor4f(0.1, 0.1, 0.1, 1.0);
            glPushMatrix();
            {
                glTranslatef(0.0, 0.8, 0);
                glScalef(0.03, 0.8, 0.03);
                glTranslatef(2.0, 0, 0);
                glCenterUnitCylinder(7);
                glTranslatef(-4.0, 0, 0);
                glCenterUnitCylinder(7);
            }
            glPopMatrix();
        }
        glPopMatrix();
    }
    glPopAttrib();
}

void cPlasmaGun::drawEffect() {
    glPushAttrib(GL_ENABLE_BIT | GL_CURRENT_BIT | GL_ALL_ATTRIB_BITS);
    {

        glDisable(GL_CULL_FACE);
        glDisable(GL_LIGHTING);
        glDisable(GL_TEXTURE_2D);
        glBlendFunc(GL_SRC_ALPHA, GL_ONE);
        glDepthMask(GL_FALSE);

        float n[16];
        glGetTransposeInverseRotationMatrix(n);

        foreachNoInc(i, shrapnelParticles) {
            cParticle* s = *i++;
            glPushMatrix();
            {
                glTranslatef(s->pos[0], s->pos[1], s->pos[2]);
                glMultMatrixf(n);
                glColor4f(0.1, 0.1, 0.6, 0.99f);
                glDisk(9, 1.9 * 0.1f);
                glColor4f(0.8, 0.8, 0.1, 0.6f);
                glDisk(7, 1.9 * 0.07f);
            }
            glPopMatrix();
        }

    }
    glPopAttrib();
}

void cPlasmaGun::drawHUD() {
    float a = 0.9;
    float b = 0.6;

    // Ammo
    float i = 0.1;
    float j = 0.9;
    float r = i + (remainingAmmo / (float) clipSize) * (j - i);
    glBegin(GL_QUADS);
    glColor4f(1, 1, 0, b);
    glVertex3f(i, j, 0.0);
    glVertex3f(i, i, 0.0);
    glColor4f(1, 0, 0, b);
    glVertex3f(r, i, 0.0);
    glVertex3f(r, j, 0.0);
    glEnd();

    // Clips
    float l = 0.1;
    float h = 0.15;
    r = i + (remainingClips / (float) depotSize) * (j - i);
    glBegin(GL_QUADS);
    glColor4f(0, 0, 0, a);
    glVertex3f(l, i + h, 0.0);
    glVertex3f(l, i, 0.0);
    glColor4f(1, 1, 1, a);
    glVertex3f(r, i, 0.0);
    glVertex3f(r, i + h, 0.0);
    glEnd();

    glBegin(GL_LINES);
    glColor4f(0, 0, 1, b);
    glVertex3f(0.25, 0.5, 0.0);
    glColor4f(1, 1, 1, a);
    glVertex3f(0.75, 0.5, 0.0);
    glEnd();
}


//--------------------------------

cSparkGun::cSparkGun() {
    clipSize = 24;
    depotSize = 3;
    remainingAmmo = clipSize;
    remainingClips = depotSize;

    if (WEAPONSOUND) {
        ALuint buffer = alutCreateBufferFromFile("data/weapon/sgun/sparkgun.wav");
        assert(buffer != AL_NONE);
        alGenSources(1, &soundSource);
        alSourcei(soundSource, AL_BUFFER, buffer);
        alSourcef(soundSource, AL_PITCH, 1.0f);
        alSourcef(soundSource, AL_GAIN, 1.0f);
        alSourcei(soundSource, AL_LOOPING, AL_FALSE);
    }
}

void cSparkGun::fire(OID target) {
    if (!ready()) return;

    if (remainingAmmo > 0) {
        remainingAmmo--;
        if (remainingAmmo == 0 && remainingClips != 0) {
            timeReloading = 7.1;
        } else {
            timeReloading = 1.9;
        }
    }

    float* source = weaponPosef;

    cParticle* s = new cParticle();
    assert(s != NULL);
    s->fuel = 1;

    float* source_12 = &source[12];
    vector_cpy(s->pos, source_12)

            float nrm[3];
    float pos2[] = {0, 0, -1};
    matrix_apply2(source, pos2);
    vector_sub(nrm, pos2, s->pos);

    vector_scale(s->vel, nrm, 30);
    shrapnelParticles.push_back(s);

    playSourceIfNotPlaying();
}

void cSparkGun::animate(float spf) {

    foreachNoInc(i, shrapnelParticles) {
        cParticle* s = *i++;
        s->pos[0] += s->vel[0] * spf;
        s->pos[1] += s->vel[1] * spf;
        s->pos[2] += s->vel[2] * spf;
        s->fuel -= spf;
        if (s->fuel < 0) {
            shrapnelParticles.remove(s);
            delete s;
        } else {
            float radius = 0.4;
            int roles = 0;
            float damage = 11;
            int damaged = this->damageByParticle(s->pos.data(), radius, roles, damage);
            if (damaged) {
                shrapnelParticles.remove(s);
                delete s;
            }
        }
    }

    timeReloading -= spf;
    if (timeReloading < 0) {
        timeReloading = 0.0f;
        if (remainingAmmo == 0 && remainingClips != 0) {
            remainingAmmo = clipSize;
            if (remainingClips > 0) remainingClips--;
        }
    }
}

void cSparkGun::drawSolid() {
    glPushAttrib(GL_ALL_ATTRIB_BITS);
    {
        glDisable(GL_CULL_FACE);
        glPushMatrix();
        {
            glMultMatrixf(weaponPosef);
            glRotatef(-90, 1, 0, 0);

            // Scale
            float scale = weaponScale;
            glScalef(scale, scale, scale);

            if (this->ready() == 0 && remainingAmmo != 0) {
            }

            glColor4f(0.3, 0.2, 0.2, 1.0);
            glPushMatrix();
            {
                glScalef(0.1, 0.14, 0.12);
                glCenterUnitBlock();
            }
            glPopMatrix();

            glEnable(GL_LIGHTING);
            glColor4f(0.1, 0.1, 0.1, 1.0);
            glPushMatrix();
            {
                glTranslatef(0.0, 0.7, 0);
                glScalef(0.1, 0.7, 0.1);
                glCenterUnitCylinder(7);
            }
            glPopMatrix();
        }
        glPopMatrix();
    }
    glPopAttrib();
}

void cSparkGun::drawEffect() {
    glPushAttrib(GL_ALL_ATTRIB_BITS);
    {

        glDisable(GL_CULL_FACE);
        glDisable(GL_LIGHTING);
        glDisable(GL_TEXTURE_2D);
        glBlendFunc(GL_SRC_ALPHA, GL_ONE);
        glDepthMask(GL_FALSE);

        float n[16];
        glGetTransposeInverseRotationMatrix(n);

        foreachNoInc(i, shrapnelParticles) {
            cParticle* s = *i++;
            glPushMatrix();
            {
                glTranslatef(s->pos[0], s->pos[1], s->pos[2]);
                glMultMatrixf(n);
                glColor4f(0.6, 0.1, 0.1, 0.99f);
                glDisk(9 + DETAIL, 5 * 0.1f + 0.006 * (rand() % 100));
                glColor4f(0.9, 0.9, 0.4, 0.99f);
                glDisk(7 + DETAIL, 5 * 0.07f + 0.006 * (rand() % 100));
            }
            glPopMatrix();
        }

    }
    glPopAttrib();
}

void cSparkGun::drawHUD() {
    float a = 0.9;
    float b = 0.6;

    // Ammo
    float i = 0.1;
    float j = 0.9;
    float r = i + (remainingAmmo / (float) clipSize) * (j - i);
    glBegin(GL_QUADS);
    glColor4f(1, 1, 0, b);
    glVertex3f(i, j, 0.0);
    glVertex3f(i, i, 0.0);
    glColor4f(1, 0, 0, b);
    glVertex3f(r, i, 0.0);
    glVertex3f(r, j, 0.0);
    glEnd();

    // Clips
    float l = 0.1;
    float h = 0.15;
    r = i + (remainingClips / (float) depotSize) * (j - i);
    glBegin(GL_QUADS);
    glColor4f(0, 0, 0, a);
    glVertex3f(l, i + h, 0.0);
    glVertex3f(l, i, 0.0);
    glColor4f(1, 1, 1, a);
    glVertex3f(r, i, 0.0);
    glVertex3f(r, i + h, 0.0);
    glEnd();

    glBegin(GL_LINES);
    glColor4f(1, 1, 0, b);
    glEnd();
}


//--------------------------------

cRaybeam::cRaybeam() {
    clipSize = 12;
    depotSize = 1;
    remainingAmmo = clipSize;
    remainingClips = depotSize;

    if (WEAPONSOUND) {
        ALuint buffer = alutCreateBufferFromFile("data/weapon/rbeam/raybeam.wav");
        assert(buffer != AL_NONE);
        alGenSources(1, &soundSource);
        alSourcei(soundSource, AL_BUFFER, buffer);
        alSourcef(soundSource, AL_PITCH, 1.0f);
        alSourcef(soundSource, AL_GAIN, 1.0f);
        alSourcei(soundSource, AL_LOOPING, AL_FALSE);
    }
};

void cRaybeam::fire(OID target) {
    if (!ready()) return;

    if (remainingAmmo > 0) {
        remainingAmmo--;
        if (remainingAmmo == 0 && remainingClips != 0) {
            timeReloading = 2.0;
        } else {
            timeReloading = 2.0;
            timeFiring = 0.7;
        }
    }

    playSourceIfNotPlaying();
}

void cRaybeam::animate(float spf) {
    timeReloading -= spf;
    if (timeReloading < 0) {
        timeReloading = 0.0f;
        if (remainingAmmo == 0 && remainingClips != 0) {
            remainingAmmo = clipSize;
            if (remainingClips > 0) remainingClips--;
        }
    }

    timeFiring -= spf;
    if (timeFiring < 0) {
        timeFiring = 0.0f;
        return;
    }

    float* source = weaponPosef;

    // Vector that points to the Mountpoint relative to the eye.
    float* mpnt = &source[12];
    //vector_print(mpnt);

    // Mountpoint-Forward Vector relative to the eye.
    float* source_8 = &source[8];
    float mfwd[3];
    vector_cpy(mfwd, source_8);
    //vector_print(mfwd);
    vector_norm(mfwd, mfwd);

    // Interpolate single points towards the "tip" of the ray.
    float offset = 80 * (1.0f - (timeFiring / 0.7));
    float worldpos[] = {0, 0, 0};
    vector_scale(worldpos, mfwd, -offset);
    vector_add(worldpos, worldpos, mpnt);
    //vector_print(worldpos);

    // Collide and Damage with single point on ray.
    {
        float radius = 0.5;
        int roles = 0;
        float damage = 25;
        int damaged = damageByParticle(worldpos, radius, roles, damage);
        // Stop ray on collision.
        if (damaged > 0) timeFiring = 0;
    }
}

void cRaybeam::drawSolid() {
    glPushAttrib(GL_ALL_ATTRIB_BITS);
    {
        glDisable(GL_CULL_FACE);
        glPushMatrix();
        {
            glMultMatrixf(weaponPosef);
            glRotatef(-90, 1, 0, 0);

            // Scale
            float scale = weaponScale;
            glScalef(scale, scale, scale);

            if (this->ready() == 0 && remainingAmmo != 0) {
            }

            glColor4f(0.3, 0.2, 0.3, 1.0);
            glPushMatrix();
            {
                glScalef(0.1, 0.14, 0.12);
                glCenterUnitBlock();
            }
            glPopMatrix();

            glEnable(GL_LIGHTING);
            glColor4f(0.1, 0.1, 0.1, 1.0);
            glPushMatrix();
            {
                glTranslatef(0.0, 0.7, 0);
                glScalef(0.03, 0.7, 0.03);
                glCenterUnitCylinder(7);
                glTranslatef(0.0, 1.0, 0);
                glScalef(2.0, 0.1, 2.0);
                glColor4f(0.4, 0.4, 0.4, 1.0);
                glCenterUnitCylinder(7);
            }
            glPopMatrix();
        }
        glPopMatrix();
    }
    glPopAttrib();
}

void cRaybeam::drawEffect() {
    if (timeFiring == 0) return;

    const float len = 10;

    struct appearance {
        float width, length;
        float icolor[4];
        float ocolor[4];
    } app[] = {
        {
            0.10, 8 * len,
            {0.3, 0.3, 0.3, 1},
            {0.3, 0.0, 0.0, 1},
        },
        {
            0.40, 8 * len,
            {0.5, 0.5, 0.5, 1},
            {0.3, 0.0, 0.0, 1},
        },
        {
            0.80, 8 * len,
            {0.9, 0.9, 0.9, 1},
            {0.3, 0.0, 0.0, 1},
        },
        {
            0.85, 8 * len,
            {1.0, 1.0, 1.0, 1},
            {0.4, 0.0, 0.0, 1},
        },
        {
            0.90, 8.2 * len,
            {1.0, 1.0, 1.0, 1},
            {0.7, 0.0, 0.0, 1},
        },
        {
            0.60, 8.4 * len,
            {0.9, 0.9, 0.9, 1},
            {0.3, 0.0, 0.0, 1},
        },
        {
            0.30, 8.5 * len,
            {0.7, 0.7, 0.7, 1},
            {0.3, 0.0, 0.0, 1},
        }
    };

    unsigned int lifeindex = (int) ((0.7 - timeFiring) * 10);
    if (lifeindex > sizeof (app) / sizeof (appearance) - 1) lifeindex = sizeof (app) / sizeof (appearance) - 1;
    if (lifeindex < 0) lifeindex = 0;

    glPushAttrib(GL_ALL_ATTRIB_BITS);
    {
        glDisable(GL_CULL_FACE);
        glDisable(GL_LIGHTING);
        glDisable(GL_TEXTURE_2D);
        glBlendFunc(GL_SRC_ALPHA, GL_ONE);
        glDepthMask(GL_FALSE);

        glPushMatrix();
        {
            glMultMatrixf(weaponPosef);

            float m[16];
            glGetFloatv(GL_MODELVIEW_MATRIX, m);
            float* mpnt = &m[12]; // Vector that points to the Mountpoint relative to the eye.
            //vector_print(mpnt);
            float* mfwd = &m[8]; // Mountpoint-Forward Vector relative to the eye.
            //vector_print(mfwd);

            float right[3];
            vector_cross(right, mpnt, mfwd);
            vector_norm(right, right);
            //vector_print(right);

            //if (mpnt[2] > 0 || mfwd[2] < 0) right[1] = -right[1]; // Correction
            //if (mpnt[1] > 0 || mfwd[1] < 0) right[1] = -right[1]; // Correction

            const float w = app[lifeindex].width;
            right[0] *= -w;
            right[1] *= -w;
            right[2] *= -w;
            glBegin(GL_TRIANGLE_STRIP);
            glColor4fv(app[lifeindex].ocolor);
            glVertex3f(-right[0], -right[1], -right[2] - 2 * w);
            glVertex3f(-right[0], -right[1], -right[2] - app[lifeindex].length);
            glColor4fv(app[lifeindex].icolor);
            glVertex3f(0, 0.0, 0);
            glVertex3f(0, 0.0, -app[lifeindex].length);
            glColor4fv(app[lifeindex].ocolor);
            glVertex3f(+right[0], +right[1], +right[2] - 2 * w);
            glVertex3f(+right[0], +right[1], +right[2] - app[lifeindex].length);
            glEnd();

        }
        glPopMatrix();
    }
    glPopAttrib();
}

void cRaybeam::drawHUD() {
    float a = 0.9;
    float b = 0.6;

    // Ammo
    float i = 0.1;
    float j = 0.9;
    float r = i + (remainingAmmo / (float) clipSize) * (j - i);
    glBegin(GL_QUADS);
    glColor4f(1, 1, 0, b);
    glVertex3f(i, j, 0.0);
    glVertex3f(i, i, 0.0);
    glColor4f(1, 0, 0, b);
    glVertex3f(r, i, 0.0);
    glVertex3f(r, j, 0.0);
    glEnd();

    // Clips
    float l = 0.1;
    float h = 0.15;
    r = i + (remainingClips / (float) depotSize) * (j - i);
    glBegin(GL_QUADS);
    glColor4f(0, 0, 0, a);
    glVertex3f(l, i + h, 0.0);
    glVertex3f(l, i, 0.0);
    glColor4f(1, 1, 1, a);
    glVertex3f(r, i, 0.0);
    glVertex3f(r, i + h, 0.0);
    glEnd();

    glBegin(GL_LINES);
    glColor4f(1, 0.3, 0.3, b);
    glVertex3f(0.25, 0.5, 0.0);
    glColor4f(1, 0.6, 0.6, a);
    glVertex3f(0.75, 0.5, 0.0);
    glEnd();
}

//--------------------------------

cHomingMissile::cHomingMissile() {
    clipSize = 4;
    depotSize = 7;
    remainingAmmo = clipSize;
    remainingClips = depotSize;

    if (WEAPONSOUND) {
        ALuint buffer = alutCreateBufferFromFile("data/weapon/hmissile/homingmissile.wav");
        assert(buffer != AL_NONE);
        alGenSources(1, &soundSource);
        alSourcei(soundSource, AL_BUFFER, buffer);
        alSourcef(soundSource, AL_PITCH, 1.0f);
        alSourcef(soundSource, AL_GAIN, 1.0f);
        alSourcei(soundSource, AL_LOOPING, AL_FALSE);
    }
}

void cHomingMissile::fire(OID target) {
    if (!ready()) return;

    if (remainingAmmo > 0) {
        remainingAmmo--;
        if (remainingAmmo == 0 && remainingClips != 0) {
            timeReloading = 2.0;
        } else {
            timeReloading = 0.8;
        }
    }

    if (target == 0) {
        target = this->weaponOwner->enemyNearby();
    }

    float* source = weaponPosef;

    cParticle* s = new cParticle();
    assert(s != NULL);
    s->target = target;
    s->fuel = 4;
    s->spawn = 0.0f;
    s->timer = 0.0f;

    float* pos = &source[12];
    vector_cpy(s->pos, pos);

    float nrm[3];
    float pos2[] = {0, 0, -1};

    // Variation of direction.
    //pos2[0] += (100-(rand()%200)) * 0.005;
    //pos2[1] += (100-(rand()%200)) * 0.005 + 0.1;
    //pos2[2] += (100-(rand()%200)) * 0.01;

    matrix_apply2(source, pos2);
    vector_sub(nrm, pos2, s->pos);

    vector_scale(s->vel, nrm, 1);
    missileParticles.push_back(s);

    playSource();
}

void cHomingMissile::animate(float spf) {

    foreachNoInc(i, missileParticles) {
        cParticle* s = *i++;

        // Update Particle from Missile Trails.
        bool removed_smoke = false;

        foreachNoInc(j, s->trail) {
            cParticle* smoke = *j++;
            smoke->fuel -= spf;
            smoke->timer += spf;
            if (smoke->fuel < 0) {
                s->trail.remove(smoke);
                removed_smoke = true;
            }
        }

        const int MAX_SMOKE = 50;
        // Add new Smoke-Particle whenever an old Smoke-Particel disappears
        // or at specified intervals.
        s->spawn -= spf;
        if (((int) s->trail.size() < MAX_SMOKE) && (s->spawn <= 0 || removed_smoke)) {
            s->spawn = 0.04;
            cParticle* smoke = new cParticle();
            assert(smoke != NULL);
            smoke->pos = s->pos;
            //vector_cpy(smoke->pos, s->pos);
            smoke->fuel = 0.9f;
            smoke->timer = 0.0f;
            s->trail.push_back(smoke);
        }

        // Now update the Missile itself
        if (s->target) {
            cObject *targetobject = cWorld::instance->mIndex[s->target];
            if (targetobject && true) {
                float tgt[] = {0, 0, 0};
                vector_cpy(tgt, targetobject->traceable->pos);
                float dir[] = {tgt[0] - s->pos[0], tgt[1] - s->pos[1], tgt[2] - s->pos[2]};
                vector_norm(dir, dir);
                vector_norm(s->vel, s->vel);
                const float alpha = 0.96;
                loop3i(s->vel[i] = (s->vel[i] * alpha + dir[i]*(1 - alpha)));
            }
        }
        vector_norm(s->vel, s->vel);
        const float speed = 30;
        s->pos[0] += speed * s->vel[0] * spf;
        s->pos[1] += speed * s->vel[1] * spf;
        s->pos[2] += speed * s->vel[2] * spf;
        s->fuel -= spf;
        s->timer += spf;

        // When fuel is empty just let the Missile and all
        // it's Smoke-Particles disappear.
        if (s->fuel < 0) {

            foreachNoInc(j, s->trail) {
                cParticle* smoke = *j++;
                s->trail.remove(smoke);
            }
            missileParticles.remove(s);
            delete s;
        } else {
            float radius = 0.25;
            int roles = 0;
            float damage = 16;
            int damaged = this->damageByParticle(s->pos.data(), radius, roles, damage);
            if (damaged) {
                missileParticles.remove(s);
                delete s;
            }
        }
    }

    timeReloading -= spf;
    if (timeReloading < 0) {
        timeReloading = 0.0f;
        if (remainingAmmo == 0 && remainingClips != 0) {
            remainingAmmo = clipSize;
            if (remainingClips > 0) remainingClips--;
        }
    }
};

void cHomingMissile::drawSolid() {
    // Missile barrels

    glPushAttrib(GL_ALL_ATTRIB_BITS);
    {
        glDisable(GL_CULL_FACE);
        glPushMatrix();
        {
            glMultMatrixf(weaponPosef);
            glRotatef(-90, 1, 0, 0);

            // Scale
            float scale = weaponScale;
            glScalef(scale, scale, scale);

            glColor4f(0.3, 0.3, 0.3, 1.0);
            glPushMatrix();
            {
                glTranslatef(0, 0.75, 0);
                glScalef(0.065, 0.75, 0.065);
                glCenterUnitCylinder(7);

            }
            glPopMatrix();

            glColor4f(0.3, 0.3, 0.2, 1.0);
            glPushMatrix();
            {
                glScalef(0.1, 0.14, 0.12);
                glCenterUnitBlock();
            }
            glPopMatrix();

            glColor4f(0.1, 0.1, 0.1, 1.0);
            if (this->ready() == 0 && remainingAmmo != 0) glRotatef(this->weaponOwner->base->seconds * 1 * 90, 0, 1, 0);
            int n = 4;

            loopi(n) {
                glPushMatrix();
                {
                    glRotatef(i * 360 / n, 0, 1, 0);
                    glTranslatef(0.075, 0.8, 0);
                    glScalef(0.06, 0.8, 0.06);
                    glCenterUnitCylinder(7);
                }
                glPopMatrix();
            }
        }
        glPopMatrix();
    }
    glPopAttrib();

    // Missiles themselves:
    glPushAttrib(GL_ALL_ATTRIB_BITS);
    {

        glDisable(GL_CULL_FACE);
        glDisable(GL_LIGHTING);
        glDisable(GL_TEXTURE_2D);

        const float size = 0.8;
        const float length = size * 1.4;
        const float radius = size * 0.2;
        const float s1 = radius * sin(0 * PI_OVER_180);
        const float c1 = radius * cos(0 * PI_OVER_180);
        const float s2 = radius * sin(120 * PI_OVER_180);
        const float c2 = radius * cos(120 * PI_OVER_180);
        const float s3 = radius * sin(240 * PI_OVER_180);
        const float c3 = radius * cos(240 * PI_OVER_180);

        float m[16];
        glGetFloatv(GL_MODELVIEW_MATRIX, m);
        glColor4f(0.7, 0.7, 0.7, 1.0);

        foreachNoInc(i, missileParticles) {
            cParticle* s = *i++;

            float r[3];
            vector_cross(r, s->vel, (&m[4]));
            float u[3];
            vector_cross(u, s->vel, r);
            float b[16];
            memset(b, 0, sizeof (float) *16);
            float* r_vec = &b[0];
            float* u_vec = &b[4];
            float* d_vec = &b[8];
            vector_cpy(r_vec, r);
            vector_cpy(u_vec, u);
            vector_cpy(d_vec, s->vel);
            b[15] = 1;
            //matrix_print(b);

            glPushMatrix();
            {
                glTranslatef(s->pos[0], s->pos[1], s->pos[2]);
                glMultMatrixf(b);
                //glNormAxis();
                glBegin(GL_TRIANGLE_FAN);
                glVertex3f(0, 0, length);
                glVertex3f(s1, c1, 0);
                glVertex3f(s2, c2, 0);
                glVertex3f(s3, c3, 0);
                glVertex3f(s1, c1, 0);
                glEnd();
            }
            glPopMatrix();
        }

    }
    glPopAttrib();
}

void cHomingMissile::drawEffect() {
    glPushAttrib(GL_ALL_ATTRIB_BITS);
    {

        glDisable(GL_CULL_FACE);
        glDisable(GL_LIGHTING);
        glDisable(GL_TEXTURE_2D);
        glDepthMask(GL_FALSE);
        glBlendFunc(GL_SRC_ALPHA, GL_ONE);

        float n[16];
        glGetTransposeInverseRotationMatrix(n);

        // Draw Missile's flare.
        glBlendFunc(GL_SRC_ALPHA, GL_ONE);

        foreachNoInc(i, missileParticles) {
            cParticle* s = *i++;

            glPushMatrix();
            {
                glTranslatef(s->pos[0], s->pos[1], s->pos[2]);
                glMultMatrixf(n);
                float size = 2.5 + 2.5 * 0.01f * (rand() % 100);
                glColor4f(0.8, 0.5, 0.1, 0.99f);
                glDisk(9 + DETAIL, size * 0.1f);
                size = 2.5 + 2.5 * 0.01f * (rand() % 100);
                glColor4f(0.9, 0.9, 0.9, 0.99f);
                glDisk(7 + DETAIL, size * 0.07f);
            }
            glPopMatrix();
        }

        // Draw Missile's trails.
        //glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
        glBlendFunc(GL_SRC_ALPHA, GL_ONE);
        glEnable(GL_TEXTURE_2D);
        glBindTexture(GL_TEXTURE_2D, sTextures[0]);

        foreachNoInc(i, missileParticles) {
            cParticle* s = *i++;

            foreachNoInc(j, s->trail) {
                cParticle* smoke = *j++;
                glPushMatrix();
                {
                    glTranslatef(smoke->pos[0], smoke->pos[1], smoke->pos[2]);
                    glMultMatrixf(n);
                    float m = 2 * smoke->fuel;
                    glColor4f(m, m * 0.8, m * 0.6, 0.05 + smoke->fuel);
                    float size = 0.1 + 0.2 * pow(1.0 + smoke->timer, 2.0);
                    glDisk(5 + DETAIL, size);
                }
                glPopMatrix();
            }
        }

    }
    glPopAttrib();
}

void cHomingMissile::drawHUD() {
    float a = 0.9;
    float b = 0.6;

    // Ammo
    float i = 0.1;
    float j = 0.9;
    float r = i + (remainingAmmo / (float) clipSize) * (j - i);
    glBegin(GL_QUADS);
    glColor4f(1, 1, 0, b);
    glVertex3f(i, j, 0.0);
    glVertex3f(i, i, 0.0);
    glColor4f(1, 0, 0, b);
    glVertex3f(r, i, 0.0);
    glVertex3f(r, j, 0.0);
    glEnd();

    // Clips
    float l = 0.1;
    float h = 0.15;
    r = i + (remainingClips / (float) depotSize) * (j - i);
    glBegin(GL_QUADS);
    glColor4f(0, 0, 0, a);
    glVertex3f(l, i + h, 0.0);
    glVertex3f(l, i, 0.0);
    glColor4f(1, 1, 1, a);
    glVertex3f(r, i, 0.0);
    glVertex3f(r, i + h, 0.0);
    glEnd();

    glBegin(GL_LINES);
    glColor4f(0.3, 1, 0.3, b);
    glVertex3f(0.4, 0.4, 0.0);
    glVertex3f(0.6, 0.6, 0.0);
    glVertex3f(0.6, 0.4, 0.0);
    glVertex3f(0.4, 0.6, 0.0);
    glVertex3f(0.5, 0.4, 0.0);
    glVertex3f(0.5, 0.6, 0.0);
    glVertex3f(0.4, 0.5, 0.0);
    glVertex3f(0.6, 0.5, 0.0);
    glEnd();
}


//--------------------------------

cExplosion::cExplosion() {
    if (EXPLOSIONSOUND) {
        unsigned buffer;
        //buffer = alutCreateBufferHelloWorld();
        buffer = alutCreateBufferFromFile("data/weapon/explosion/explosion.wav");
        assert(buffer != AL_NONE);
        alGenSources(1, &soundSource);
        if (alGetError() != AL_FALSE || !alIsSource(soundSource)) throw "sorry, could not create audio source for explosion";
        alSourcei(soundSource, AL_BUFFER, buffer);
        alSourcef(soundSource, AL_PITCH, 1.0f);
        alSourcef(soundSource, AL_GAIN, 10.0f);
        alSourcei(soundSource, AL_LOOPING, AL_FALSE);
    }
}

void cExplosion::fire(OID target) {
    if (!ready()) return;
    timeReloading = 3.8;

    // Shoot several shrapnels around.

    loopi(11) {

        float* source = weaponPosef;

        cParticle* s = new cParticle();
        assert(s != NULL);
        s->target = target;
        s->fuel = 5;
        s->spawn = 0.0f;
        s->timer = 0.0f;

        s->pos[0] = source[12];
        s->pos[1] = source[13];
        s->pos[2] = source[14];

        float nrm[3];
        float pos2[] = {0, 0, -1};

        pos2[0] = (100 - (rand() % 200)) * 0.01;
        pos2[1] = 0.4 + (100 - (rand() % 200)) * 0.01;
        pos2[2] = (100 - (rand() % 200)) * 0.01;

        matrix_apply2(source, pos2);
        vector_sub(nrm, pos2, s->pos);

        vector_scale(s->vel, nrm, 1);
        missileParticles.push_back(s);
    } // for

    playSourceIfNotPlaying();
}

void cExplosion::animate(float spf) {

    foreachNoInc(i, missileParticles) {
        cParticle* s = *i++;

        // Update Smoke from Missile Trails.
        bool removed_smoke = false;

        foreachNoInc(j, s->trail) {
            cParticle* smoke = *j++;
            smoke->fuel -= spf;
            smoke->timer += spf;
            if (smoke->fuel < 0) {
                s->trail.remove(smoke);
                removed_smoke = true;
            }
        }

        const int MAX_SMOKE = 100;
        // Add new Smoke-Particle whenever an old Smoke-Particel disappears
        // or at specified intervals.
        s->spawn -= spf;
        if (((int) s->trail.size() < MAX_SMOKE) && (s->spawn <= 0 || removed_smoke)) {
            s->spawn = 0.12;
            cParticle* smoke = new cParticle();
            assert(smoke != NULL);
            smoke->pos = s->pos;
            //vector_cpy(smoke->pos, s->pos);
            smoke->fuel = 1.2f;
            smoke->timer = 0.0f;
            s->trail.push_back(smoke);
        }

        const float alpha = 0.992f;
        loop3i(s->vel[i] *= alpha);

        const float speed = 5;
        s->pos[0] += speed * s->vel[0] * spf;
        s->pos[1] += speed * s->vel[1] * spf;
        s->pos[2] += speed * s->vel[2] * spf;
        s->fuel -= spf;
        s->timer += spf;

        // When fuel is empty just let the Missile and all
        // it's Smoke-Particles disappear.
        if (s->fuel < 0) {

            foreachNoInc(j, s->trail) {
                cParticle* smoke = *j++;
                s->trail.remove(smoke);
            }
            missileParticles.remove(s);
            delete s;
        } else {
            float radius = 0.25;
            int roles = 0;
            float damage = 10;
            int damaged = this->damageByParticle(s->pos.data(), radius, roles, damage);
            if (damaged) {
                missileParticles.remove(s);
                delete s;
            }
        }
    }

    timeReloading -= spf;
    if (timeReloading < 0) timeReloading = 0.0f;
};

void cExplosion::drawSolid() {
    return; // !!

    glPushAttrib(GL_ALL_ATTRIB_BITS);
    {

        glDisable(GL_CULL_FACE);
        glDisable(GL_LIGHTING);
        glDisable(GL_TEXTURE_2D);

        glColor4f(0.7, 0.7, 0.7, 1.0);

        const float size = 0.8 * this->weaponScale;
        const float length = size * 0.4;
        const float radius = size * 0.8;
        const float s1 = radius * sin(0 * PI_OVER_180);
        const float c1 = radius * cos(0 * PI_OVER_180);
        const float s2 = radius * sin(120 * PI_OVER_180);
        const float c2 = radius * cos(120 * PI_OVER_180);
        const float s3 = radius * sin(240 * PI_OVER_180);
        const float c3 = radius * cos(240 * PI_OVER_180);

        float m[16];
        glGetFloatv(GL_MODELVIEW_MATRIX, m);

        foreachNoInc(i, missileParticles) {
            cParticle* s = *i++;

            float r[3];
            vector_cross(r, s->vel, (&m[4]));
            float u[3];
            vector_cross(u, s->vel, r);
            float b[16];
            memset(b, 0, sizeof (float) *16);
            float* r_vec = &b[0];
            float* u_vec = &b[4];
            float* d_vec = &b[8];
            vector_cpy(r_vec, r);
            vector_cpy(u_vec, u);
            vector_cpy(d_vec, s->vel);
            b[15] = 1;
            //matrix_print(b);

            glPushMatrix();
            {
                glTranslatef(s->pos[0], s->pos[1], s->pos[2]);
                glMultMatrixf(b);
                //glNormAxis();
                glBegin(GL_TRIANGLE_FAN);
                glVertex3f(0, 0, length);
                glVertex3f(s1, c1, 0);
                glVertex3f(s2, c2, 0);
                glVertex3f(s3, c3, 0);
                glVertex3f(s1, c1, 0);
                glEnd();
            }
            glPopMatrix();
        }

    }
    glPopAttrib();
}

void cExplosion::drawEffect() {
    glPushAttrib(GL_ALL_ATTRIB_BITS);
    {

        glDisable(GL_CULL_FACE);
        glDisable(GL_LIGHTING);
        glDisable(GL_TEXTURE_2D);
        glDepthMask(GL_FALSE);

        float n[16];
        glGetTransposeInverseRotationMatrix(n);

        /*
        // Draw Missile's flare.
        glBlendFunc(GL_SRC_ALPHA, GL_ONE);

        foreachNoInc(i, mMissiles) {
            Particle* s = *i++;

            glPushMatrix();
            {
                glTranslatef(s->pos[0], s->pos[1], s->pos[2]);
                glMultMatrixf(n);
                float size = 2.5 + 2.5 * 0.01f * (rand() % 100);
                glColor4f(0.8, 0.5, 0.1, 0.99f);
                glDisk(9 + DETAIL, size * 0.1f);
                size = 2.5 + 2.5 * 0.01f * (rand() % 100);
                glColor4f(0.9, 0.9, 0.9, 0.99f);
                glDisk(7 + DETAIL, size * 0.07f);
            }
            glPopMatrix();
        }
         */

        // Draw Missile's trails.
        //glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
        glBlendFunc(GL_SRC_ALPHA, GL_ONE);
        glEnable(GL_TEXTURE_2D);
        glBindTexture(GL_TEXTURE_2D, sTextures[0]);

        foreachNoInc(i, missileParticles) {
            cParticle* s = *i++;

            foreachNoInc(j, s->trail) {
                cParticle* smoke = *j++;
                glPushMatrix();
                {
                    glTranslatef(smoke->pos[0], smoke->pos[1], smoke->pos[2]);
                    glMultMatrixf(n);
                    float m = 2 * smoke->fuel;
                    glColor4f(m, m * 0.8, m * 0.6, 0.07 + smoke->fuel);
                    float size = 0.1 + 0.3 * pow(1.0 + smoke->timer, 3.5);
                    glDisk(5 + DETAIL, size);
                }
                glPopMatrix();
            }
        }

    }
    glPopAttrib();
}


